
/*
hawk (Herv's awk) v1.0
Copyright (C) by Denis Valois && Cdric Llorens, 2008.
denis.valois@laposte.fr, cedric.llorens@wanadoo.fr

This software is a computer program whose purpose is to perform
pattern scanning and processing.  This is governed by a given
template containing patterns structured in regular expressions.

This software is governed by the CeCILL license under French law and
abiding by the rules of distribution of free software.  You can  use,
modify and/or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info".

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability.

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or
data to be ensured and,  more generally, to use and operate it in the
same conditions as regards security.

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <strings.h>
#include <assert.h>

#include "atom.h"
#include "xmalloc.h"


struct atom_t *atom_new(struct flags_t flags, unsigned long lineno, char *str)
{
	struct atom_t *ap;
	int cflags, status;
	size_t err_lg;
	static char err_buf[BUFSIZ];

	assert(str != NULL);

	ap = xmalloc(sizeof(struct atom_t));

	ap->ref_cnt = 1;

	ap->flags.type = flags.type;
	ap->flags.kase = flags.kase;
	ap->flags.line = flags.line;

	ap->lineno = lineno;
	ap->str = xstrdup(str);

	if (flags.type == flag_f)
	{
		ap->preg = NULL;
	}
	else
	{
		ap->preg = xmalloc(sizeof(regex_t));

		cflags = REG_NEWLINE;
		if (flags.type == flag_e) cflags |= REG_EXTENDED;
		if (flags.kase == flag_i) cflags |= REG_ICASE;
		/* REG_NOSUB never set even if flags.line == flag_x */

		if ((status = regcomp(ap->preg, str, cflags)) != 0)
		{
			err_lg = regerror(status, ap->preg, err_buf, sizeof(err_buf));
			assert(err_lg < sizeof(err_buf));
			if (lineno != 0)
			{
				fprintf(stderr, "error at line %lu\n", lineno);
			}
			else
			{
				fprintf(stderr, "error at line <unknown>\n");
			}
			fprintf(stderr, "%s\n", str);
			fprintf(stderr, "%s\n", err_buf);
			exit(EXIT_FAILURE);
		}
	}

	return ap;
}


#if defined(__linux__) || defined(__svr4__)
char *strcasestr(char *haystack, char *needle)
{
	char *cp, *lim;

	lim = haystack + strlen(haystack) - strlen(needle);
	for (cp = haystack; cp <= lim; cp++)
	{
		if (strcasecmp(cp, needle) == 0) return cp;
	}
	return NULL;
}
#endif


int atom_match(char *line, size_t lg, struct atom_t *ap)
{
	int ok;

	assert(line != NULL);
	assert(ap != NULL);
	assert(ap->str != NULL);

	if (ap->flags.type != flag_f)
	{
		/* compare wrt a regex */
		assert(ap->preg != NULL);

		if (ap->flags.line == flag_x)
		{
			/* rcx rix */
			regmatch_t pmatch;

			if (regexec(ap->preg, line, 1, &pmatch, 0) == 0)
			{
				ok = (pmatch.rm_so == 0 && pmatch.rm_eo ==  (int) lg);
			}
			else
			{
				ok = 0;
			}
		}
		else
		{
			/* rcs ris */
			ok = (regexec(ap->preg, line, 0, NULL, 0) == 0);
		}
	}
	else
	{
		/* compare wrt a fixed string */
		assert(ap->preg == NULL);

		if (ap->flags.kase == flag_i)
		{
			if (ap->flags.line == flag_x)
			{
				/* fix */
				ok = (strcasecmp(line, ap->str) == 0);
			}
			else
			{
				/* fis */
				ok = (strcasestr(line, ap->str) != NULL);
			}
		}
		else
		{
			if (ap->flags.line == flag_x)
			{
				/* fcx */
				ok = (strcmp(line, ap->str) == 0);
			}
			else
			{
				/* fcs */
				ok = (strstr(line, ap->str) != NULL);
			}
		}
	}

	return ok;
}


void atom_free(struct atom_t **app)
{
	struct atom_t *ap;

	assert(app != NULL);
	ap = *app;
	assert(ap != NULL);

	if (--(ap->ref_cnt) == 0)
	{
		assert(ap->str != NULL);
		free( ap->str );

		if (ap->flags.type == flag_f)
		{
			assert(ap->preg == NULL);
		}
		else
		{
			assert(ap->preg != NULL);
			regfree( ap->preg );
			free( ap->preg );
		}

		free( ap );
		(*app) = NULL;
	}

	return;
}
