
/*
hawk (Herv's awk) v1.0
Copyright (C) by Denis Valois && Cdric Llorens, 2008.
denis.valois@laposte.fr, cedric.llorens@wanadoo.fr

This software is a computer program whose purpose is to perform
pattern scanning and processing.  This is governed by a given
template containing patterns structured in regular expressions.

This software is governed by the CeCILL license under French law and
abiding by the rules of distribution of free software.  You can  use,
modify and/or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info".

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability.

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or
data to be ensured and,  more generally, to use and operate it in the
same conditions as regards security.

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.
*/

#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "avl.h"
#include "xmalloc.h"

struct avl_t *avl_search2(struct avl_t **root, char *key, int *inserted, int *height);

struct avl_t *avl_remove2(struct avl_t **root, char *key, int *height);
struct avl_t *avl_perm_pred(struct avl_t **root, struct avl_t **pred, int *height);
void avl_bal_left(struct avl_t **root, int *height);
void avl_bal_right(struct avl_t **root, int *height);


struct avl_t *avl_search(struct avl_t **root, char *key, int *inserted)
{
	int height;

	assert(root != NULL);
	assert(key != NULL);

	return avl_search2(root, key, inserted, &height);
}


struct avl_t *avl_remove(struct avl_t **root, char *key)
{
	int height;

	assert(root != NULL);
	assert(key != NULL);

	return avl_remove2(root, key, &height);
}


struct avl_t *avl_search2(struct avl_t **root, char *key, int *inserted, int *height)
{
	struct avl_t *p, *p1, *p2, *r;
	int cmp;

	p = *root;
	if (p != NULL)
	{
		assert(p->key != NULL);
		cmp = strcmp(key, p->key);

		if (cmp < 0)
		{
			r = avl_search2( &(p->left), key, inserted, height );
			if (*height)
			{
				/* left branch has grown 1 level */
				assert(inserted != NULL && *inserted);
				switch (p->bal)
				{
					case bal_left:	/* rebalance */
						p1 = p->left;
						assert(p1 != NULL);
						if (p1->bal == bal_left)
						{
							/* single LL rotation */
							p->left = p1->right;
							p1->right = p;
							p->bal = bal_equal;
							p = p1;
						}
						else
						{
							/* double LR rotation */
							p2 = p1->right;
							assert(p2 != NULL);
							p1->right = p2->left;
							p2->left = p1;
							p->left = p2->right;
							p2->right = p;
							switch (p2->bal)
							{
								case bal_left:
									p->bal = bal_right;
									p1->bal = bal_equal;
									break;
								case bal_equal:
									p->bal = bal_equal;
									p1->bal = bal_equal;
									break;
								case bal_right:
									p->bal = bal_equal;
									p1->bal = bal_left;
									break;
							}
							p = p2;
						}
						p->bal = bal_equal;
						*height = 0;
						break;

					case bal_equal:
						p->bal = bal_left;
						break;

					case bal_right:
						p->bal = bal_equal;
						*height = 0;
						break;
				}
			}
		}
		else if (cmp > 0)
		{
			r = avl_search2( &(p->right), key, inserted, height );
			if (*height)
			{
				/* right branch has grown 1 level */
				assert(inserted != NULL && *inserted);
				switch (p->bal)
				{
					case bal_left:
						p->bal = bal_equal;
						*height = 0;
						break;

					case bal_equal:
						p->bal = bal_right;
						break;

					case bal_right:	/* rebalance */
						p1 = p->right;
						assert(p1 != NULL);
						if (p1->bal == bal_right)
						{
							/* single RR rotation */
							p->right = p1->left;
							p1->left = p;
							p->bal = bal_equal;
							p = p1;
						}
						else
						{
							/* double RL rotation */
							p2 = p1->left;
							assert(p2 != NULL);
							p1->left = p2->right;
							p2->right = p1;
							p->right = p2->left;
							p2->left = p;
							switch (p2->bal)
							{
								case bal_left:
									p->bal = bal_equal;
									p1->bal = bal_right;
									break;
								case bal_equal:
									p->bal = bal_equal;
									p1->bal = bal_equal;
									break;
								case bal_right:
									p->bal = bal_left;
									p1->bal = bal_equal;
									break;
							}
							p = p2;
						}
						p->bal = bal_equal;
						*height = 0;
						break;
				}
			}
		}
		else /* cmp == 0 */
		{
			r = p;
			if (inserted != NULL) *inserted = 0;
			*height = 0;
		}
	}
	else /* p == NULL */
	{
		if (inserted != NULL)
		{
			/* create and insert new node */
			p = xmalloc(sizeof(struct avl_t));
			p->key = xstrdup(key);
			p->u.idtable = NULL;
			p->left = NULL;
			p->right = NULL;
			p->bal = bal_equal;
			r = p;
			*inserted = 1;
			*height = 1;
		}
		else
		{
			/* returns "not found" */
			r = NULL;
			*height = 0;
		}
	}

	assert((inserted != NULL) ? (r != NULL && (*inserted || ! *height)) : (*root == p && ! *height));

	*root = p;
	return r;
}


struct avl_t *avl_remove2(struct avl_t **root, char *key, int *height)
{
	struct avl_t *p, *r;
	int cmp;

	p = *root;
	if (p != NULL)
	{
		cmp = strcmp(key, p->key);
		if (cmp < 0)
		{
			r = avl_remove2( &(p->left), key, height );
			if (*height) avl_bal_left(root, height);
		}
		else if (cmp > 0)
		{
			r = avl_remove2( &(p->right), key, height );
			if (*height) avl_bal_right(root, height);
		}
		else /* cmp == 0 */
		{
			if (p->right == NULL)
			{
				r = p;
				*root = p->left;
				*height = 1;
			}
			else if (p->left == NULL)
			{
				r = p;
				*root = p->right;
				*height = 1;
			}
			else
			{
				r = avl_perm_pred(root, &(p->left), height);
				if (*height) avl_bal_left(root, height);
			}
		}
	}
	else
	{
		r = NULL;
		*height = 0;
	}

	return r;
}


struct avl_t *avl_perm_pred(struct avl_t **root, struct avl_t **pred, int *height)
{
	struct avl_t *p, *r;
	char *key;

	p = *pred;
	assert(p != NULL);
	if (p->right != NULL)
	{
		r = avl_perm_pred(root, &(p->right), height);
		if (*height) avl_bal_right(pred, height);
	}
	else
	{
		key = (*root)->key;
		(*root)->key = p->key;
		p->key = key;
		/* the following test is constant; thus optimizable */
		if (sizeof(p->u) == sizeof(p->u.vstr))
		{
			char *t;

			t = (*root)->u.vstr;
			(*root)->u.vstr = p->u.vstr;
			p->u.vstr = t;
		}
		else
		{
			long int t;

			assert(sizeof(p->u) == sizeof(p->u.vint));
			t = (*root)->u.vint;
			(*root)->u.vint = p->u.vint;
			p->u.vint = t;
		}
		r = p;
		(*pred) = p->left;
		*height = 1;
	}

	return r;
}


void avl_bal_left(struct avl_t **root, int *height)
{
	struct avl_t *p, *p1, *p2;
	enum avl_bal_t b1, b2;

	p = *root;
	switch (p->bal)
	{
		case bal_left:
			p->bal = bal_equal;
			break;

		case bal_equal:
			p->bal = bal_right;
			*height = 0;
			break;

		case bal_right:	/* rebalance */
			p1 = p->right;  b1 = p1->bal;
			switch (b1)
			{
				case bal_left:
					/* double RL rotation */
					p2 = p1->left;  b2 = p2->bal;
					p1->left = p2->right;
					p2->right = p1;
					p->right = p2->left;
					p2->left = p;
					switch (b2)
					{
						case bal_left:
							p->bal = bal_equal;
							p1->bal = bal_right;
							break;
						case bal_equal:
							p->bal = bal_equal;
							p1->bal = bal_equal;
							break;
						case bal_right:
							p->bal = bal_left;
							p1->bal = bal_equal;
							break;
					}
					p = p2;
					p->bal = bal_equal;
					break;

				case bal_equal:
				case bal_right:
					/* single RR rotation */
					p->right = p1->left;
					p1->left = p;
					if (b1 == bal_equal)
					{
						p->bal = bal_right;
						p1->bal = bal_left;
						*height = 0;
					}
					else
					{
						p->bal = bal_equal;
						p1->bal = bal_equal;
					}
					p = p1;
					break;
			}
			break;
	}

	*root = p;
	return;
}


void avl_bal_right(struct avl_t **root, int *height)
{
	struct avl_t *p, *p1, *p2;
	enum avl_bal_t b1, b2;

	p = *root;
	switch (p->bal)
	{
		case bal_left:	/* rebalance */
			p1 = p->left;  b1 = p1->bal;
			switch (b1)
			{
				case bal_left:
				case bal_equal:
					/* single LL rotation */
					p->left = p1->right;
					p1->right = p;
					if (b1 == bal_equal)
					{
						p->bal = bal_left;
						p1->bal = bal_right;
						*height = 0;
					}
					else
					{
						p->bal = bal_equal;
						p1->bal = bal_equal;
					}
					p = p1;
					break;

				case bal_right:
					/* double LR rotation */
					p2 = p1->right;  b2 = p2->bal;
					p1->right = p2->left;
					p2->left = p1;
					p->left = p2->right;
					p2->right = p;
					switch (b2)
					{
						case bal_left:
							p->bal = bal_right;
							p1->bal = bal_equal;
							break;
						case bal_equal:
							p->bal = bal_equal;
							p1->bal = bal_equal;
							break;
						case bal_right:
							p->bal = bal_equal;
							p1->bal = bal_left;
							break;
					}
					p = p2;
					p->bal = bal_equal;
					break;
			}
			break;

		case bal_equal:
			p->bal = bal_left;
			*height = 0;
			break;

		case bal_right:
			p->bal = bal_equal;
			break;
	}

	*root = p;
	return;
}


void avl_delete(struct avl_t **root, enum vtype_t vtype)
{
	struct avl_t *ptr;

	assert(root != NULL);
	ptr = *root;

	if (ptr != NULL)
	{
		avl_delete( &(ptr->left), vtype );
		avl_delete( &(ptr->right), vtype );

		assert(ptr->key != NULL);
		free(ptr->key);

		/* used only for tab:  don't delete ptr->u.idtable */
		if (vtype == t_str)
		{
			assert(ptr->u.vstr != NULL);
			free(ptr->u.vstr);
		}

		free(ptr);
		*root = NULL;
	}

	return;
}


void avl_free(struct avl_t *root)
{
	if (root != NULL)
	{
		avl_free( root->left );
		avl_free( root->right );

		assert(root->key != NULL);
		free( root->key );

		idtable_free( root->u.idtable );

		free( root );
	}
	return;
}
