
/*
hawk (Herv's awk) v1.0
Copyright (C) by Denis Valois && Cdric Llorens, 2008.
denis.valois@laposte.fr, cedric.llorens@wanadoo.fr

This software is a computer program whose purpose is to perform
pattern scanning and processing.  This is governed by a given
template containing patterns structured in regular expressions.

This software is governed by the CeCILL license under French law and
abiding by the rules of distribution of free software.  You can  use,
modify and/or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info".

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability.

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or
data to be ensured and,  more generally, to use and operate it in the
same conditions as regards security.

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.
*/

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <assert.h>

#include "dump.h"


char *image_vtype[] =
{
	"(str)",
	"(int)"
};

char *image_top[] =
{
	"not",
	"star",
	"plus",
	"maybe",
	"or",
	"and",
	"concat",
	"leaf"
};

char *image_stat[] =
{
	"stat list",
	"expression",
	"if",
	"while",
	"do",
	"for",
	"forall",
	"break",
	"continue",
	"delete",
	"exit",
	"accept",
	"reject"
};

char *image_expr[] =
{
	",",
	"? :",
	"||",
	"&&",
	"|",
	"&",
	"!",
	"~",
	"+ unary",
	"- unary",
	"^",
	"==",
	"!=",
	"<=",
	">=",
	"<",
	">",
	"<<",
	">>",
	"+ binary",
	"- binary",
	"*",
	"/",
	"%",
	"++ suffix",
	"-- suffix",
	"++ prefix",
	"-- prefix",
	"array index",
	"=",
	"+=",
	"-=",
	"*=",
	"/=",
	"%=",
	"^=",
	"|=",
	"&=",
	"<<=",
	">>=",
	"identifier",
	"constant\t(int)",
	"constant\t(str)"
};

char *image_state[] =
{
	"1-line (fast evaluation)",
	"epsilon",
	"conjunction init",
	"conjunction",
	"accept"
};

void image_str(char *cp);

void dump_tab(struct avl_t *root, enum vtype_t vtype);


void dump_avl(struct avl_t *root)
{
	if (root != NULL)
	{
		dump_avl( root->left );

		fprintf(DUMP_FP, "avl\t%p\n", (void *) root);
		fprintf(DUMP_FP, "\tkey\t\t");
		image_str(root->key);
		fprintf(DUMP_FP, "\n\tidtable\t\t%p\n", (void *) root->u.idtable);
		dump_idtable( root->u.idtable );

		dump_avl( root->right );
	}

	return;
}


void dump_pattern(struct pattern_t *root)
{
	if (root != NULL)
	{
		fprintf(DUMP_FP, "pattern %p\n", (void *) root);
		fprintf(DUMP_FP, "\t%s\n", image_top[root->top]);
		fprintf(DUMP_FP, "\tnullable\t%d\n\tpattern1\t%d\n",
		        root->nullable, root->pattern1);
		fprintf(DUMP_FP, "\taction\t\t%p\n", (void *) root->action);
		if (root->top == top_leaf)
		{
			fprintf(DUMP_FP, "\tatom\t\t%p\n", (void *) root->u.ap);
		}
		else
		{
			fprintf(DUMP_FP, "\tchild1\t\t%p\n\tchild2\t\t%p\n",
			        (void *) root->u.n.child1, (void *) root->u.n.child2);
		}

		dump_statement( root->action );

		if (root->top == top_leaf)
		{
			dump_atom( root->u.ap );
		}
		else
		{
			dump_pattern( root->u.n.child1 );
			dump_pattern( root->u.n.child2 );
		}
	}

	return;
}


void dump_statement(struct statement_t *root)
{
	struct stat_list_t *sl;

	if (root != NULL)
	{
		fprintf(DUMP_FP, "statement %p\n", (void *) root);

		fprintf(DUMP_FP, "\t%s", image_stat[root->st_type]);
		switch (root->st_type)
		{
			case st_list:
				fprintf(DUMP_FP, "\t%p", (void *) root->u.p_list);
				break;
			case st_expression:
				fprintf(DUMP_FP, "\t%p", (void *) root->u.p_expression);
				break;
			default:
				/* noop */
				break;
		}
		putc('\n', DUMP_FP);

		switch (root->st_type)
		{
			case st_list:
				for (sl = root->u.p_list; sl != NULL; sl = sl->next)
				{
					fprintf(DUMP_FP, "\t\t%p\n", (void *) sl->sp);
				}

				for (sl = root->u.p_list; sl != NULL; sl = sl->next)
				{
					dump_statement( sl->sp );
				}
				break;

			case st_expression:
				dump_expression( root->u.p_expression );
				break;

			case st_if:
				fprintf(DUMP_FP, "\t\texpr\t%p\n",
				        (void *) root->u.p_if.expr);
				fprintf(DUMP_FP, "\t\tthen\t%p\n",
				        (void *) root->u.p_if.p_then);
				fprintf(DUMP_FP, "\t\telse\t%p\n",
				        (void *) root->u.p_if.p_else);

				dump_expression( root->u.p_if.expr );
				dump_statement( root->u.p_if.p_then );
				dump_statement( root->u.p_if.p_else );
				break;

			case st_while:
			case st_do:
				fprintf(DUMP_FP, "\t\texpr\t%p\n",
				        (void *) root->u.p_while.expr);
				fprintf(DUMP_FP, "\t\tbody\t%p\n",
				        (void *) root->u.p_while.p_body);

				dump_expression( root->u.p_while.expr );
				dump_statement( root->u.p_while.p_body );
				break;

			case st_for:
				fprintf(DUMP_FP, "\t\texpr1\t%p\n",
				        (void *) root->u.p_for.expr1);
				fprintf(DUMP_FP, "\t\texpr2\t%p\n",
				        (void *) root->u.p_for.expr2);
				fprintf(DUMP_FP, "\t\texpr3\t%p\n",
				        (void *) root->u.p_for.expr3);
				fprintf(DUMP_FP, "\t\tbody\t%p\n",
				        (void *) root->u.p_for.p_body);

				dump_expression( root->u.p_for.expr1 );
				dump_expression( root->u.p_for.expr2 );
				dump_expression( root->u.p_for.expr3 );
				dump_statement( root->u.p_for.p_body );
				break;

			case st_forall:
				fprintf(DUMP_FP, "\t\tarray\t%p\n",
				        (void *) root->u.p_forall.p_array);
				fprintf(DUMP_FP, "\t\tindex\t%p\n",
				        (void *) root->u.p_forall.p_index);
				fprintf(DUMP_FP, "\t\tvalue\t%p\n",
				        (void *) root->u.p_forall.p_value);
				fprintf(DUMP_FP, "\t\tbody\t%p\n",
				        (void *) root->u.p_forall.p_body);

				dump_statement( root->u.p_forall.p_body );
				break;

			case st_break:
			case st_continue:
			case st_accept:
			case st_reject:
				break;

			case st_delete:
			case st_exit:
				fprintf(DUMP_FP, "\t\texpr\t%p\n",
				        (void *) root->u.p_delete.expr1);

				dump_expression( root->u.p_delete.expr1 );
				break;
		}
	}

	return;
}


void dump_expression(struct expression_t *root)
{
	unsigned int i;

	if (root != NULL)
	{
		fprintf(DUMP_FP, "expression %p\n", (void *) root);
		fprintf(DUMP_FP, "\t%s", image_expr[root->expr_node]);
		switch (root->expr_node)
		{
			case n_id:
				fprintf(DUMP_FP, "\t%p", (void *) root->u.id);
				break;

			case n_intcst:
				fprintf(DUMP_FP, "\n\t\t%ld", root->u.vint);
				break;

			case n_strcst:
				fprintf(DUMP_FP, "\n\t\t");
				image_str(root->u.vstr);
				break;

			default:
				/* noop */
				break;
		}
		putc('\n', DUMP_FP);

		for (i = 0; i < root->opc; i++)
		{
			fprintf(DUMP_FP, "\targ %2u\t\t%p\n", i, (void *) root->opv[i]);
		}

		for (i = 0; i < root->opc; i++)
		{
			dump_expression( root->opv[i] );
		}
	}

	return;
}


void dump_idtable(struct idtable_t *root)
{
	if (root != NULL)
	{
		fprintf(DUMP_FP, "idtable\t%p\n", (void *) root);
		switch (root->id_kind)
		{
			case id_macro:
				fprintf(DUMP_FP, "\tmacro pat\t%p\n", (void *) root->u.s1.v_pat);
				fprintf(DUMP_FP, "\tmacro stat\t%p\n", (void *) root->u.s1.v_stat);
				dump_pattern( root->u.s1.v_pat );
				dump_statement( root->u.s1.v_stat );
				break;

			case id_var:
				fprintf(DUMP_FP, "\tvar\t%s\n", image_vtype[root->u.v_var.vtype]);
				fprintf(DUMP_FP, "\t\t");
				switch (root->u.v_var.vtype)
				{
					case t_str:
						image_str(root->u.v_var.u.vstr);
						break;
					case t_int:
						fprintf(DUMP_FP, "%ld", root->u.v_var.u.vint);
						break;
				}
				putc('\n', DUMP_FP);
				break;

			case id_fct:
				fprintf(DUMP_FP, "\tfct\t\t%s\n", image_vtype[root->u.s2.vtype]);
				break;

			case id_tab:
				fprintf(DUMP_FP, "\ttab\t\t%s\n", image_vtype[root->u.v_tab.vtype]);
				dump_tab( root->u.v_tab.root, root->u.v_tab.vtype );
				break;
		}
	}

	return;
}


void dump_atom(struct atom_t *root)
{
	if (root != NULL)
	{
		fprintf(DUMP_FP, "atom\t%p\n", (void *) root);

		fprintf(DUMP_FP, "\tflags\t\t[");
		switch (root->flags.type)
		{
			case flag_f:
				fprintf(DUMP_FP, "f");
				break;
			case flag_r:
				fprintf(DUMP_FP, "r");
				break;
			case flag_e:
				fprintf(DUMP_FP, "e");
				break;
		}
		switch (root->flags.kase)
		{
			case flag_c:
				fprintf(DUMP_FP, "c");
				break;
			case flag_i:
				fprintf(DUMP_FP, "i");
				break;
		}
		switch (root->flags.line)
		{
			case flag_s:
				fprintf(DUMP_FP, "s");
				break;
			case flag_x:
				fprintf(DUMP_FP, "x");
				break;
		}
		fprintf(DUMP_FP, "]\n");

		fprintf(DUMP_FP, "\tline\t%lu\t", root->lineno);
		image_str(root->str);
		putc('\n', DUMP_FP);
	}

	return;
}


void image_str(char *cp)
{
	int c;

	assert(cp != NULL);

	putc('"', DUMP_FP);
	while ((c = *cp++) != '\0')
	{
		switch (c)
		{
			case '\\':
			case '"':
				putc('\\', DUMP_FP);
				putc(c, DUMP_FP);
				break;

			case '\t':
				putc('\\', DUMP_FP);
				putc('t', DUMP_FP);
				break;

			case '\v':
				putc('\\', DUMP_FP);
				putc('v', DUMP_FP);
				break;

			case '\f':
				putc('\\', DUMP_FP);
				putc('f', DUMP_FP);
				break;

			case '\n':
				putc('\\', DUMP_FP);
				putc('n', DUMP_FP);
				break;

			case '\r':
				putc('\\', DUMP_FP);
				putc('r', DUMP_FP);
				break;

			default:
				if (isprint(c))
				{
					putc(c, DUMP_FP);
				}
				else
				{
					fprintf(DUMP_FP, "\\0x%02x", c);
				}
				break;
		}
	}
	putc('"', DUMP_FP);

	return;
}


void dump_tab(struct avl_t *root, enum vtype_t vtype)
{
	if (root != NULL)
	{
		dump_tab( root->left, vtype );

		fprintf(DUMP_FP, "\t\t[");
		image_str(root->key);
		fprintf(DUMP_FP, "] = ");

		switch (vtype)
		{
			case t_str:
				image_str(root->u.vstr);
				break;

			case t_int:
				fprintf(DUMP_FP, "%ld", root->u.vint);
				break;
		}
		putc('\n', DUMP_FP);

		dump_tab( root->right, vtype );
	}

	return;
}


void dump_nfa(struct nfa_t *nfa)
{
	unsigned long i;

	assert(nfa != NULL);
	assert(nfa->start != NULL);
	assert(nfa->accept != NULL);
	assert(nfa->state_siz != 0);
	assert(nfa->state != NULL);

	fprintf(DUMP_FP, "\tstart\t%p\t%lu\n", (void *) nfa->start, nfa->start->state_id);
	fprintf(DUMP_FP, "\taccept\t%p\t%lu\n", (void *) nfa->accept, nfa->accept->state_id);
	fprintf(DUMP_FP, "\tstates\t%lu\n", nfa->nb_states);

	for (i = 0; i < nfa->nb_states; i++)
	{
		dump_state( nfa->state[i] );
	}

	return;
}


void dump_ablock(struct ablock_t *ablock)
{
	unsigned long id;
	struct ablock_conj_t *conj;

	assert(ablock != NULL);

	fprintf(DUMP_FP, "ablock\t%p\t%s\n", (void *) ablock, (ablock == main_ablock) ? "main" : "sub");
	fprintf(DUMP_FP, "\tnfa\t\t%p\n", (void *) ablock->nfa);

	fprintf(DUMP_FP, "\tst_pres\t\t[");
	for (id = 0; id < ablock->st_pres.set_size; id++)
	{
		if (ablock->st_pres.set[id]) fprintf(DUMP_FP, " %lu", id);
	}
	fprintf(DUMP_FP, " ]\n");

	for (conj = ablock->head; conj != NULL; conj = conj->next)
	{
		fprintf(DUMP_FP, "\tconj\t\t%p\n", (void *) conj);
	}

	for (conj = ablock->head; conj != NULL; conj = conj->next)
	{
		dump_conj( ablock, conj );
	}

	return;
}


void dump_conj(struct ablock_t *ablock, struct ablock_conj_t *conj)
{
	assert(ablock != NULL);
	assert(conj != NULL);

	fprintf(DUMP_FP, "conj\t%p\n", (void *) conj);
	fprintf(DUMP_FP, "\tablock\t\t%p\n", (void *) ablock);
	fprintf(DUMP_FP, "\tid\t\t%lu\n", conj->conj_id);
	fprintf(DUMP_FP, "\tchild1\t\t%p\n", (void *) conj->child1);
	fprintf(DUMP_FP, "\tchild2\t\t%p\n", (void *) conj->child2);

	dump_ablock( conj->child1 );
	dump_ablock( conj->child2 );

	return;
}


void dump_state(struct nfa_state_t *state)
{
	assert(state != NULL);

	fprintf(DUMP_FP, "\tstate\t%p\t%lu\n", (void *) state, state->state_id);
	fprintf(DUMP_FP, "\t\t%s\n", image_state[state->state_type]);

	switch (state->state_type)
	{
		case state_1line:
			fprintf(DUMP_FP, "\t\tpattern\t%p\n", (void *) state->u.pattern);
			break;

		case state_epsilon:
			fprintf(DUMP_FP, "\t\taction\t%p\n", (void *) state->u.action);
			break;

		case state_conj_init:
		case state_conj:
			fprintf(DUMP_FP, "\t\tnfa1\t%p\n", (void *) state->u.s.nfa1);
			fprintf(DUMP_FP, "\t\tnfa2\t%p\n", (void *) state->u.s.nfa2);
			break;

		case state_accept:
			/* noop */
			break;
	}

	fprintf(DUMP_FP, "\t\tnext1\t%p", (void *) state->next1);
	if (state->next1 != NULL) fprintf(DUMP_FP, "\t%lu", state->next1->state_id);
	putc('\n', DUMP_FP);
	fprintf(DUMP_FP, "\t\tnext2\t%p", (void *) state->next2);
	if (state->next2 != NULL) fprintf(DUMP_FP, "\t%lu", state->next2->state_id);
	putc('\n', DUMP_FP);

	return;
}
