
/*
hawk (Herv's awk) v1.0
Copyright (C) by Denis Valois && Cdric Llorens, 2008.
denis.valois@laposte.fr, cedric.llorens@wanadoo.fr

This software is a computer program whose purpose is to perform
pattern scanning and processing.  This is governed by a given
template containing patterns structured in regular expressions.

This software is governed by the CeCILL license under French law and
abiding by the rules of distribution of free software.  You can  use,
modify and/or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info".

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability.

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or
data to be ensured and,  more generally, to use and operate it in the
same conditions as regards security.

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.
*/

#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>

#include "expression.h"
#include "xmalloc.h"
#include "io.h"
#include "hawk.h"
#include "avl.h"
#include "atom.h"

FILE *filep(long int fn);
void expr_eval_opv(struct expression_t *e);
char *expr_eval_field(struct expression_t *e);
long int expr_eval_nf(struct expression_t *e);
char *expr_eval_strcat(struct expression_t *e);
char *expr_eval_substr(char *s, long int i, long int l);
long int expr_eval_split(struct expression_t *e);
char *expr_eval_sprintf(struct expression_t *e, unsigned long nargs);
long int expr_eval_match(struct expression_t *e);


void expr_eval_fct(struct expression_t *e, struct evalue_t *v)
{
	struct idtable_t *id;
	unsigned int lg;
	char *cp;
	int ret, sig;

	assert(e != NULL);
	assert(v != NULL);
	assert(e->expr_node == n_id);
	id = e->u.id;
	assert(id != NULL);
	assert(id->id_kind == id_fct);

	switch (id->u.s2.v_fct)
	{
		case f_fflush:
			assert(e->opc == 1);
			v->vtype = t_int;
			v->u.vint = fflush( filep(expr_eval_int(e->opv[0])) );
			break;

		case f_field:
			assert(e->opc == 1 || e->opc == 2 || e->opc == 3);
			v->vtype = t_str;
			v->u.vstr = expr_eval_field(e);
			break;

		case f_filename:
			assert(e->opc == 0);
			v->vtype = t_str;
			v->u.vstr = xstrdup(io_filename);
			break;

		case f_fprintf:
			assert(e->opc >= 2);
			expr_eval_opv(e);
			/* file stream is the last parameter */
			assert(e->vop[e->opc - 1].vtype == t_int);
			cp = expr_eval_sprintf(e, e->opc - 1);
			assert(cp != NULL);
			v->vtype = t_int;
			v->u.vint = fprintf( filep(e->vop[e->opc - 1].u.vint), "%s", cp );
			free(cp);
			break;

		case f_index:
			/* leftmost character is at index 1 */
			assert(e->opc == 2);
			expr_eval( e->opv[0], &(e->vop[0]) );
			expr_eval( e->opv[1], &(e->vop[1]) );
			assert(e->vop[0].vtype == t_str);
			assert(e->vop[1].vtype == t_str);
			cp = strstr( e->vop[0].u.vstr, e->vop[1].u.vstr );
			v->vtype = t_int;
			v->u.vint = (cp != NULL) ? (cp - e->vop[0].u.vstr + 1) : 0;
			break;

		case f_inttostr:
			assert(e->opc == 1);
			v->vtype = t_str;
			v->u.vstr = xstrdup(" ");
			v->u.vstr[0] = (char) expr_eval_int(e->opv[0]);
			break;

		case f_length:
			assert(e->opc == 1);
			cp = expr_eval_str(e->opv[0]);
			v->vtype = t_int;
			v->u.vint = strlen( cp );
			free( cp );
			break;

		case f_line:
			assert(e->opc == 0);
			v->vtype = t_str;
			v->u.vstr = xstrdup(io_line);
			break;

		case f_lineno:
			assert(e->opc == 0);
			v->vtype = t_int;
			v->u.vint = io_lineno;
			break;

		case f_nf:
			assert(e->opc == 0 || e->opc == 1 || e->opc == 2);
			v->vtype = t_int;
			v->u.vint = expr_eval_nf(e);
			break;

		case f_match:
			assert(e->opc == 2 || e->opc == 3);
			v->vtype = t_int;
			v->u.vint = expr_eval_match(e);
			break;

		case f_printf:
			assert(e->opc >= 1);
			expr_eval_opv(e);
			cp = expr_eval_sprintf(e, e->opc);
			assert(cp != NULL);
			v->vtype = t_int;
			v->u.vint = printf("%s", cp );
			free(cp);
			break;

		case f_split:
			assert(e->opc == 2 || e->opc == 3);
			v->vtype = t_int;
			v->u.vint = expr_eval_split(e);
			break;

		case f_sprintf:
			assert(e->opc >= 1);
			expr_eval_opv(e);
			v->vtype = t_str;
			v->u.vstr = expr_eval_sprintf(e, e->opc);
			break;

		case f_stderr:
			assert(e->opc == 0);
			v->vtype = t_int;
			v->u.vint = fileno(stderr);
			break;

		case f_stdout:
			assert(e->opc == 0);
			v->vtype = t_int;
			v->u.vint = fileno(stdout);
			break;

		case f_strcat:
			assert(e->opc >= 2);
			v->vtype = t_str;
			v->u.vstr = expr_eval_strcat(e);
			break;

		case f_strtoint:
			assert(e->opc == 1);
			cp = expr_eval_str(e->opv[0]);
			v->vtype = t_int;
			v->u.vint = (long int) ((unsigned char) *cp);
			free( cp );
			break;

		case f_strtonum:
			assert(e->opc == 1);
			cp = expr_eval_str(e->opv[0]);
			v->vtype = t_int;
			v->u.vint = strtol( cp, NULL, 0 );
			free( cp );
			break;

		case f_substr:
			/* leftmost character is at index 1 */
			assert(e->opc == 2 || e->opc == 3);
			expr_eval_opv(e);
			assert(e->vop[0].vtype == t_str);
			assert(e->vop[1].vtype == t_int);
			if (e->opc == 2)
			{
				lg = strlen(e->vop[0].u.vstr);
			}
			else
			{
				assert(e->vop[2].vtype == t_int);
				lg = e->vop[2].u.vint;
			}
			v->vtype = t_str;
			v->u.vstr = expr_eval_substr(e->vop[0].u.vstr,
			                             e->vop[1].u.vint - 1,
			                             lg);
			break;

		case f_system:
			assert(e->opc == 1);
			cp = expr_eval_str(e->opv[0]);
			ret = system( cp );
			sig = WTERMSIG(ret);
			/* handle signal from system */
			if (WIFSIGNALED(ret) && (sig == SIGINT || sig == SIGQUIT))
			{
				raise(sig);
			}
			v->vtype = t_int;
			v->u.vint = (long int) ret;
			free( cp );
			break;

		case f_tolower:
			assert(e->opc == 1);
			v->vtype = t_str;
			v->u.vstr = expr_eval_str(e->opv[0]);
			for (cp = v->u.vstr; *cp != '\0'; cp++)
			{
				*cp = tolower(*cp);
			}
			break;

		case f_toupper:
			assert(e->opc == 1);
			v->vtype = t_str;
			v->u.vstr = expr_eval_str(e->opv[0]);
			for (cp = v->u.vstr; *cp != '\0'; cp++)
			{
				*cp = toupper(*cp);
			}
			break;
	}

	return;
}


FILE *filep(long int fn)
{
	FILE *fp = NULL;

	if (fn == fileno(stdout))
		fp = stdout;
	else if (fn == fileno(stderr))
		fp = stderr;
	else
	{
		runtime_error("bad IO descriptor");
	}

	return fp;
}


void expr_eval_opv(struct expression_t *e)
{
	unsigned int i;

	assert(e != NULL);
	assert(e->opc !=  0);

	for (i = 0; i < e->opc; i++)
	{
		assert(e->opv[i] != NULL);
		expr_eval(e->opv[i], &(e->vop[i]));
	}

	return;
}


char *expr_eval_field(struct expression_t *e)
{
	char *s, *fs, *cp, *r;
	long int f, i;
	int charset[256] = { 0 };

	assert(e != NULL);
	assert(e->opc == 1 || e->opc == 2 || e->opc == 3);

	f = expr_eval_int( e->opv[0] );

	if (e->opc >= 2)
	{
		s = expr_eval_str( e->opv[1] );
	}
	else
	{
		s = xstrdup(io_line);
	}

	if (e->opc == 3)
	{
		fs = expr_eval_str( e->opv[2] );
		for (cp = fs; *cp != '\0'; cp++) charset[ (unsigned char) *cp ] = 1;
		free(fs);
	}
	else
	{
		assert(pred_fs != NULL);
		assert(pred_fs->id_kind == id_var);
		assert(pred_fs->u.v_var.vtype == t_str);
		assert(pred_fs->u.v_var.u.vstr != NULL);
		for (cp = pred_fs->u.v_var.u.vstr; *cp != '\0'; cp++) charset[(unsigned char) *cp] = 1;
	}

	if (f > 0)
	{
		cp = s;

		for (i = 1; i < f && *cp != '\0'; i++)
		{
			while (*cp != '\0' && charset[(unsigned char) *cp]) cp++;
			/* cp at the beginning of $i */

			while (*cp != '\0' && ! charset[(unsigned char) *cp]) cp++;
			/* cp just after $i */
		}
		/* cp just after $(f-1) */

		while (*cp != '\0' && charset[(unsigned char) *cp]) cp++;
		/* cp just at the beginning of $f */
		r = cp;

		while (*cp != '\0' && ! charset[(unsigned char) *cp]) cp++;
		/* cp just after $f */
		*cp = '\0';

	}
	else if (f == 0)
	{
		r = s;
	}
	else /* f < 0 */
	{
		r = "";
	}

	cp = xstrdup(r);
	free(s);

	return cp;
}


long int expr_eval_nf(struct expression_t *e)
{
	long int r;
	char *s, *fs, *cp;
	int charset[256] = { 0 };

	assert(e != NULL);
	assert(e->opc == 0 || e->opc == 1 || e->opc == 2);

	if (e->opc != 0)
	{
		s = expr_eval_str( e->opv[0] );
	}
	else
	{
		s = xstrdup(io_line);
	}

	if (e->opc == 2)
	{
		fs = expr_eval_str( e->opv[1] );
		for (cp = fs; *cp != '\0'; cp++) charset[ (unsigned char) *cp ] = 1;
		free(fs);
	}
	else
	{
		assert(pred_fs != NULL);
		assert(pred_fs->id_kind == id_var);
		assert(pred_fs->u.v_var.vtype == t_str);
		assert(pred_fs->u.v_var.u.vstr != NULL);
		for (cp = pred_fs->u.v_var.u.vstr; *cp != '\0'; cp++) charset[(unsigned char) *cp] = 1;
	}

	for (cp = s; *cp != '\0' && charset[(unsigned char) *cp]; cp++) /* noop */;
	/* cp at the beginning of $1 */

	for (r = 0; *cp != '\0'; r++)
	{
		while (*cp != '\0' && ! charset[(unsigned char) *cp]) cp++;
		/* cp just after $(r+1) */

		while (*cp != '\0' && charset[(unsigned char) *cp]) cp++;
	}

	free(s);
	return r;
}


char *expr_eval_strcat(struct expression_t *e)
{
	char *buf;
	unsigned int i, lg;

	assert(e != NULL);
	assert(e->opc >= 2);

	lg = 0;
	for (i = 0; i < e->opc; i++)
	{
		expr_eval( e->opv[i], &(e->vop[i]) );
		assert(e->vop[i].vtype == t_str);
		lg += strlen(e->vop[i].u.vstr);
	}

	buf = xmalloc(lg + 1);
	buf[0] = '\0';

	for (i = 0; i < e->opc; i++)
	{
		strcat(buf, e->vop[i].u.vstr);
	}

	return buf;
}


char *expr_eval_substr(char *s, long int i, long int l)
{
	long int lg, eos;

	assert(s != NULL);
	lg = strlen(s);

	if (i < 0 || i >= lg || l <= 0)
	{
		/* generate empty string */
		s[0] = '\0';
		i = 0;
	}
	else
	{
		eos = i + l;  /* end of string position */
		if (eos < lg) s[eos] = '\0';
	}

	return xstrdup(s + i);
}


long int expr_eval_split(struct expression_t *e)
{
	long int nf;
	char *str, *cp, *fieldp, *fs;
	struct idtable_t *id;
	int blanks, inserted;
	struct avl_t *np;

	assert(e != NULL);
	assert(e->opc == 2 || e->opc == 3);

	str = expr_eval_str( e->opv[0] );

	assert(e->opv[1]->expr_node == n_id);
	id = e->opv[1]->u.id;
	assert(id != NULL);
	assert(id->id_kind == id_tab);
	assert(id->u.v_tab.vtype == t_str);
	avl_delete( &(id->u.v_tab.root), t_str );

	if (e->opc == 3)
	{
		fs = expr_eval_str( e->opv[2] );
	}
	else
	{
		assert(pred_fs != NULL);
		assert(pred_fs->id_kind == id_var);
		assert(pred_fs->u.v_var.vtype == t_str);
		assert(pred_fs->u.v_var.u.vstr != NULL);
		fs = xstrdup(pred_fs->u.v_var.u.vstr);
	}

	nf = 0;
	blanks = 1;
	fieldp = NULL;
	for (cp = str; *cp != '\0'; cp++)
	{
		if (strchr(fs, *cp) != NULL)
		{
			if (! blanks)
			{
				*cp = '\0';
				assert(fieldp != NULL);
				np = avl_search( &(id->u.v_tab.root), image_int(nf), &inserted);
				assert(inserted);
				assert(np != NULL);
				np->u.vstr = xstrdup(fieldp);
				blanks = 1;
			}
		}
		else
		{
			if (blanks)
			{
				fieldp = cp;
				nf++;
				blanks = 0;
			}
		}
	}
	if (! blanks)
	{
		assert(fieldp != NULL);
		np = avl_search( &(id->u.v_tab.root), image_int(nf), &inserted);
		assert(inserted);
		assert(np != NULL);
		np->u.vstr = xstrdup(fieldp);
	}

	free(str);
	free(fs);
	return nf;
}


char *expr_eval_sprintf(struct expression_t *e, unsigned long nargs)
{
	int r;
	char *buf;
	unsigned long bufsiz;

	assert(e != NULL);
	bufsiz = BUF_INIT;
	buf = xmalloc(bufsiz);
	r = expr_sprintf(buf, bufsiz, e, nargs);

	while (r == -1 || ((unsigned long) r) >= bufsiz)
	{
		bufsiz += BUF_INCR;
		buf = xrealloc(buf, bufsiz);
		r = expr_sprintf(buf, bufsiz, e, nargs);
	}

	buf = xrealloc(buf, strlen(buf) + 1);
	return buf;
}


long int expr_eval_match(struct expression_t *e)
{
	char *text_str, *pat_str, *flags_str, *cp;
	int type_given, kase_given, line_given, matching;
	static char *flag_error = "flag error in function match";
	struct flags_t flags;
	struct atom_t *ap;

	assert(e != NULL);
	assert(e->opc == 2 || e->opc == 3);

	text_str = expr_eval_str( e->opv[0] );
	pat_str = expr_eval_str( e->opv[1] );

	flags.type = DEFAULT_FLAG_TYPE;
	flags.kase = DEFAULT_FLAG_KASE;
	flags.line = DEFAULT_FLAG_LINE;

	if (e->opc == 3)
	{
		flags_str = expr_eval_str( e->opv[2] );
		type_given = 0;
		kase_given = 0;
		line_given = 0;

		for (cp = flags_str; *cp != '\0'; cp++)
		{
			switch (*cp)
			{
				case 'f':
				case 'F':
				case 'r':
				case 'R':
				case 'e':
				case 'E':
					if (type_given) runtime_error(flag_error);
					type_given = 1;
					break;

				case 'c':
				case 'C':
				case 'i':
				case 'I':
					if (kase_given) runtime_error(flag_error);
					kase_given = 1;
					break;

				case 's':
				case 'S':
				case 'x':
				case 'X':
					if (line_given) runtime_error(flag_error);
					line_given = 1;
					break;

				default:
					runtime_error(flag_error);
					break;
			}

			switch (*cp)
			{
				case 'f':
				case 'F':
					flags.type = flag_f;
					break;

				case 'r':
				case 'R':
					flags.type = flag_r;
					break;

				case 'e':
				case 'E':
					flags.type = flag_e;
					break;

				case 'c':
				case 'C':
					flags.kase = flag_c;
					break;

				case 'i':
				case 'I':
					flags.kase = flag_i;
					break;

				case 's':
				case 'S':
					flags.line = flag_s;
					break;

				case 'x':
				case 'X':
					flags.line = flag_x;
					break;

				default:
					/* should never reach this point */
					assert(0);
					break;
			}
		}

		free( flags_str );
	}

	ap = atom_new(flags, 0, pat_str);
	matching = atom_match(text_str, strlen(text_str), ap);

	atom_free( &ap );
	free( pat_str );
	free( text_str );

	return (long int) matching;
}
