
/*
hawk (Herv's awk) v1.0
Copyright (C) by Denis Valois && Cdric Llorens, 2008.
denis.valois@laposte.fr, cedric.llorens@wanadoo.fr

This software is a computer program whose purpose is to perform
pattern scanning and processing.  This is governed by a given
template containing patterns structured in regular expressions.

This software is governed by the CeCILL license under French law and
abiding by the rules of distribution of free software.  You can  use,
modify and/or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info".

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability.

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or
data to be ensured and,  more generally, to use and operate it in the
same conditions as regards security.

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>

#include "expression.h"
#include "hawk.h"
#include "xmalloc.h"

#define is_lvalue(ptr)	( ((ptr) != NULL) && \
			  (((ptr)->expr_node == n_id) || ((ptr)->expr_node == n_index)) )

long int *addr_int(struct expression_t *e);
char **addr_str(struct expression_t *e);
struct avl_t *addr_tab(struct expression_t *e, int create);


struct expression_t *expr_new(enum expr_node_t expr_node)
{
	struct expression_t *ep;

	ep = xmalloc(sizeof(struct expression_t));
	ep->expr_node = expr_node;

	switch (expr_node)
	{
		case n_cond:		/* ?: */
			expr_init(ep, 3);
			break;

		case n_comma:		/* , */
		case n_logor:		/* || */
		case n_logand:		/* && */
		case n_bitor:		/* | */
		case n_bitand:		/* & */
		case n_xor:		/* ^ */
		case n_eq:		/* == */
		case n_ne:		/* != */
		case n_le:		/* <= */
		case n_ge:		/* >= */
		case n_lt:		/* < */
		case n_gt:		/* > */
		case n_lshift:		/* << */
		case n_rshift:		/* >> */
		case n_add:		/* + binary */
		case n_sub:		/* - binary */
		case n_mul:		/* * */
		case n_div:		/* / */
		case n_mod:		/* % */
		case n_index:		/* [] */
		case n_assignment:	/* = */
		case n_ainc:		/* += */
		case n_adec:		/* -= */
		case n_amul:		/* *= */
		case n_adiv:		/* /= */
		case n_amod:		/* %= */
		case n_axor:		/* ^= */
		case n_aor:		/* |= */
		case n_aand:		/* &= */
		case n_lshft:		/* <<= */
		case n_rshft:		/* >>= */
			expr_init(ep, 2);
			break;

		case n_plus:		/* + unary */
		case n_minus:		/* - unary */
		case n_not:		/* ! */
		case n_compl:		/* ~ */
		case n_incrs:		/* ++ suffix */
		case n_decrs:		/* -- suffix */
		case n_incrp:		/* ++ prefix */
		case n_decrp:		/* -- prefix */
			expr_init(ep, 1);
			break;

		case n_id:
			ep->u.id = NULL;
			expr_init(ep, 0);
			break;

		case n_intcst:
			ep->u.vint = 0;
			expr_init(ep, 0);
			break;

		case n_strcst:
			ep->u.vstr = NULL;
			expr_init(ep, 0);
			break;
	}

	return ep;
}


void expr_init(struct expression_t *e, unsigned int opc)
{
	unsigned int i;

	assert(e != NULL);

	e->opc = opc;
	if (opc != 0)
	{
		e->opv = xmalloc(opc * sizeof(struct expression_t *));
		for (i = 0; i < opc; i++) e->opv[i] = NULL;
		e->vop = xmalloc(opc * sizeof(struct evalue_t));
	}
	else
	{
		e->opv = NULL;
		e->vop = NULL;
	}

	return;
}


void expr_free(struct expression_t *root)
{
	unsigned int i;

	if (root != NULL)
	{
		if (root->opc != 0)
		{
			assert(root->opv != NULL);
			assert(root->vop != NULL);

			for (i = 0; i < root->opc; i++)
			{
				assert(root->opv[i] != NULL);
				expr_free( root->opv[i] );
			}

			free(root->opv);
			free(root->vop);
		}

		if (root->expr_node == n_strcst)
		{
			assert(root->u.vstr != NULL);
			free(root->u.vstr);
		}

		free(root);
	}

	return;
}


char *expr_eval_str(struct expression_t *e)
{
	struct evalue_t r;

	assert(e != NULL);
	assert(e->vtype == t_str);
	expr_eval(e, &r);

	assert(r.vtype == t_str);
	return r.u.vstr;
}


long int expr_eval_int(struct expression_t *e)
{
	struct evalue_t r;

	assert(e != NULL);
	assert(e->vtype == t_int);
	expr_eval(e, &r);

	assert(r.vtype == t_int);
	return r.u.vint;
}


void expr_eval(struct expression_t *e, struct evalue_t *evalue)
{
	struct evalue_t v;
	long int *intptr;
	char **strptr;
	struct avl_t *tabptr;
	struct idtable_t *id;
	unsigned int i;
	static char *div0 = "division by 0";

	assert(e != NULL);
	for (i = 0; i < e->opc; i++)
	{
		e->vop[i].vtype = t_int;
		e->vop[i].u.vint = 0;
	}

	switch (e->expr_node)
	{
		case n_comma:		/* , */
			assert(e->opc == 2);
			expr_eval( e->opv[0], &(e->vop[0]) );
			expr_eval( e->opv[1], &(e->vop[1]) );
			v.vtype = e->vtype;
			switch (v.vtype)
			{
				case t_str:
					v.u.vstr = xstrdup(e->vop[1].u.vstr);
					break;
				case t_int:
					v.u.vint = e->vop[1].u.vint;
					break;
			}
			break;

		case n_cond:		/* ?: */
			assert(e->opc == 3);
			v.vtype = e->vtype;
			switch (v.vtype)
			{
				case t_str:
					v.u.vstr = ( expr_eval_int(e->opv[0]) ?
					             expr_eval_str(e->opv[1]) :
					             expr_eval_str(e->opv[2]) );
					break;
				case t_int:
					v.u.vint = ( expr_eval_int(e->opv[0]) ?
					             expr_eval_int(e->opv[1]) :
					             expr_eval_int(e->opv[2]) );
					break;
			}
			break;

		case n_logor:		/* || */
			assert(e->vtype == t_int);
			assert(e->opc == 2);
			v.vtype = t_int;
			v.u.vint = ( expr_eval_int(e->opv[0]) || expr_eval_int(e->opv[1]) );
			break;

		case n_logand:		/* && */
			assert(e->vtype == t_int);
			assert(e->opc == 2);
			v.vtype = t_int;
			v.u.vint = ( expr_eval_int(e->opv[0]) && expr_eval_int(e->opv[1]) );
			break;

		case n_bitor:		/* | */
			assert(e->vtype == t_int);
			assert(e->opc == 2);
			v.vtype = t_int;
			v.u.vint = ( expr_eval_int(e->opv[0]) | expr_eval_int(e->opv[1]) );
			break;

		case n_bitand:		/* & */
			assert(e->vtype == t_int);
			assert(e->opc == 2);
			v.vtype = t_int;
			v.u.vint = ( expr_eval_int(e->opv[0]) & expr_eval_int(e->opv[1]) );
			break;

		case n_plus:		/* + unary */
			assert(e->vtype == t_int);
			assert(e->opc == 1);
			v.vtype = t_int;
			v.u.vint = ( + expr_eval_int(e->opv[0]) );
			break;

		case n_minus:		/* - unary */
			assert(e->vtype == t_int);
			assert(e->opc == 1);
			v.vtype = t_int;
			v.u.vint = ( - expr_eval_int(e->opv[0]) );
			break;

		case n_not:		/* ! */
			assert(e->vtype == t_int);
			assert(e->opc == 1);
			v.vtype = t_int;
			v.u.vint = ( ! expr_eval_int(e->opv[0]) );
			break;

		case n_compl:		/* ~ */
			assert(e->vtype == t_int);
			assert(e->opc == 1);
			v.vtype = t_int;
			v.u.vint = ( ~ expr_eval_int(e->opv[0]) );
			break;

		case n_xor:		/* ^ */
			assert(e->vtype == t_int);
			assert(e->opc == 2);
			v.vtype = t_int;
			v.u.vint = ( expr_eval_int(e->opv[0]) ^ expr_eval_int(e->opv[1]) );
			break;

		case n_eq:		/* == */
			assert(e->vtype == t_int);
			assert(e->opc == 2);
			v.vtype = t_int;
			expr_eval( e->opv[0], &(e->vop[0]) );
			expr_eval( e->opv[1], &(e->vop[1]) );
			assert(e->vop[0].vtype == e->vop[1].vtype);
			switch (e->vop[0].vtype)
			{
				case t_str:
					v.u.vint = ( strcmp(e->vop[0].u.vstr, e->vop[1].u.vstr) == 0 );
					break;
				case t_int:
					v.u.vint = ( e->vop[0].u.vint == e->vop[1].u.vint );
					break;
			}
			break;

		case n_ne:		/* != */
			assert(e->vtype == t_int);
			assert(e->opc == 2);
			v.vtype = t_int;
			expr_eval( e->opv[0], &(e->vop[0]) );
			expr_eval( e->opv[1], &(e->vop[1]) );
			assert(e->vop[0].vtype == e->vop[1].vtype);
			switch (e->vop[0].vtype)
			{
				case t_str:
					v.u.vint = ( strcmp(e->vop[0].u.vstr, e->vop[1].u.vstr) != 0 );
					break;
				case t_int:
					v.u.vint = ( e->vop[0].u.vint != e->vop[1].u.vint );
					break;
			}
			break;

		case n_le:		/* <= */
			assert(e->vtype == t_int);
			assert(e->opc == 2);
			v.vtype = t_int;
			expr_eval( e->opv[0], &(e->vop[0]) );
			expr_eval( e->opv[1], &(e->vop[1]) );
			assert(e->vop[0].vtype == e->vop[1].vtype);
			switch (e->vop[0].vtype)
			{
				case t_str:
					v.u.vint = ( strcmp(e->vop[0].u.vstr, e->vop[1].u.vstr) <= 0 );
					break;
				case t_int:
					v.u.vint = ( e->vop[0].u.vint <= e->vop[1].u.vint );
					break;
			}
			break;

		case n_ge:		/* >= */
			assert(e->vtype == t_int);
			assert(e->opc == 2);
			v.vtype = t_int;
			expr_eval( e->opv[0], &(e->vop[0]) );
			expr_eval( e->opv[1], &(e->vop[1]) );
			assert(e->vop[0].vtype == e->vop[1].vtype);
			switch (e->vop[0].vtype)
			{
				case t_str:
					v.u.vint = ( strcmp(e->vop[0].u.vstr, e->vop[1].u.vstr) >= 0 );
					break;
				case t_int:
					v.u.vint = ( e->vop[0].u.vint >= e->vop[1].u.vint );
					break;
			}
			break;

		case n_lt:		/* < */
			assert(e->vtype == t_int);
			assert(e->opc == 2);
			v.vtype = t_int;
			expr_eval( e->opv[0], &(e->vop[0]) );
			expr_eval( e->opv[1], &(e->vop[1]) );
			assert(e->vop[0].vtype == e->vop[1].vtype);
			switch (e->vop[0].vtype)
			{
				case t_str:
					v.u.vint = ( strcmp(e->vop[0].u.vstr, e->vop[1].u.vstr) < 0 );
					break;
				case t_int:
					v.u.vint = ( e->vop[0].u.vint < e->vop[1].u.vint );
					break;
			}
			break;

		case n_gt:		/* > */
			assert(e->vtype == t_int);
			assert(e->opc == 2);
			v.vtype = t_int;
			expr_eval( e->opv[0], &(e->vop[0]) );
			expr_eval( e->opv[1], &(e->vop[1]) );
			assert(e->vop[0].vtype == e->vop[1].vtype);
			switch (e->vop[0].vtype)
			{
				case t_str:
					v.u.vint = ( strcmp(e->vop[0].u.vstr, e->vop[1].u.vstr) > 0 );
					break;
				case t_int:
					v.u.vint = ( e->vop[0].u.vint > e->vop[1].u.vint );
					break;
			}
			break;

		case n_lshift:		/* << */
			assert(e->vtype == t_int);
			assert(e->opc == 2);
			v.vtype = t_int;
			v.u.vint = ( expr_eval_int(e->opv[0]) << expr_eval_int(e->opv[1]) );
			break;

		case n_rshift:		/* >> */
			assert(e->vtype == t_int);
			assert(e->opc == 2);
			v.vtype = t_int;
			v.u.vint = ( expr_eval_int(e->opv[0]) >> expr_eval_int(e->opv[1]) );
			break;

		case n_add:		/* + binary */
			assert(e->vtype == t_int);
			assert(e->opc == 2);
			v.vtype = t_int;
			v.u.vint = ( expr_eval_int(e->opv[0]) + expr_eval_int(e->opv[1]) );
			break;

		case n_sub:		/* - binary */
			assert(e->vtype == t_int);
			assert(e->opc == 2);
			v.vtype = t_int;
			v.u.vint = ( expr_eval_int(e->opv[0]) - expr_eval_int(e->opv[1]) );
			break;

		case n_mul:		/* * */
			assert(e->vtype == t_int);
			assert(e->opc == 2);
			v.vtype = t_int;
			v.u.vint = ( expr_eval_int(e->opv[0]) * expr_eval_int(e->opv[1]) );
			break;

		case n_div:		/* / */
			assert(e->vtype == t_int);
			assert(e->opc == 2);
			v.vtype = t_int;
			expr_eval( e->opv[0], &(e->vop[0]) );
			expr_eval( e->opv[1], &(e->vop[1]) );
			assert(e->vop[0].vtype == t_int);
			assert(e->vop[1].vtype == t_int);
			if (e->vop[1].u.vint == 0) runtime_error(div0);
			v.u.vint = ( e->vop[0].u.vint / e->vop[1].u.vint );
			break;

		case n_mod:		/* % */
			assert(e->vtype == t_int);
			assert(e->opc == 2);
			v.vtype = t_int;
			expr_eval( e->opv[0], &(e->vop[0]) );
			expr_eval( e->opv[1], &(e->vop[1]) );
			assert(e->vop[0].vtype == t_int);
			assert(e->vop[1].vtype == t_int);
			if (e->vop[1].u.vint == 0) runtime_error(div0);
			v.u.vint = ( e->vop[0].u.vint % e->vop[1].u.vint );
			break;

		case n_incrs:		/* ++ suffix */
			assert(e->vtype == t_int);
			assert(e->opc == 1);
			assert(is_lvalue(e->opv[0]));
			intptr = addr_int(e->opv[0]);
			v.vtype = t_int;
			v.u.vint = (*intptr)++;
			break;

		case n_decrs:		/* -- suffix */
			assert(e->vtype == t_int);
			assert(e->opc == 1);
			assert(is_lvalue(e->opv[0]));
			intptr = addr_int(e->opv[0]);
			v.vtype = t_int;
			v.u.vint = (*intptr)--;
			break;

		case n_incrp:		/* ++ prefix */
			assert(e->vtype == t_int);
			assert(e->opc == 1);
			assert(is_lvalue(e->opv[0]));
			intptr = addr_int(e->opv[0]);
			v.vtype = t_int;
			v.u.vint = ++(*intptr);
			break;

		case n_decrp:		/* -- prefix */
			assert(e->vtype == t_int);
			assert(e->opc == 1);
			assert(is_lvalue(e->opv[0]));
			intptr = addr_int(e->opv[0]);
			v.vtype = t_int;
			v.u.vint = --(*intptr);
			break;

		case n_index:		/* [] */
			tabptr = addr_tab(e, 0);
			if (tabptr != NULL)
			{
				v.vtype = e->vtype;
				switch (e->vtype)
				{
					case t_str:
						assert(tabptr->u.vstr != NULL);
						v.u.vstr = xstrdup(tabptr->u.vstr);
						break;
					case t_int:
						v.u.vint = tabptr->u.vint;
						break;
				}
			}
			else
			{
				evalue_init(&v, e->vtype, NULL);
			}
			break;

		case n_assignment:	/* = */
			assert(e->opc == 2);
			assert(is_lvalue(e->opv[0]));
			expr_eval( e->opv[1], &(e->vop[1]) );
			switch (e->vop[1].vtype)
			{
				case t_str:
					strptr = addr_str(e->opv[0]);
					assert(*strptr != NULL);
					free(*strptr);
					assert(e->vop[1].u.vstr != NULL);
					(*strptr) = xstrdup(e->vop[1].u.vstr);
					v.vtype = t_str;
					v.u.vstr = xstrdup(*strptr);
					break;
				case t_int:
					intptr = addr_int(e->opv[0]);
					v.vtype = t_int;
					v.u.vint = ( (*intptr) = e->vop[1].u.vint );
					break;
			}
			break;

		case n_ainc:		/* += */
			assert(e->vtype == t_int);
			assert(e->opc == 2);
			assert(is_lvalue(e->opv[0]));
			intptr = addr_int(e->opv[0]);
			v.vtype = t_int;
			v.u.vint = ( (*intptr) += expr_eval_int(e->opv[1]) );
			break;

		case n_adec:		/* -= */
			assert(e->vtype == t_int);
			assert(e->opc == 2);
			assert(is_lvalue(e->opv[0]));
			intptr = addr_int(e->opv[0]);
			v.vtype = t_int;
			v.u.vint = ( (*intptr) -= expr_eval_int(e->opv[1]) );
			break;

		case n_amul:		/* *= */
			assert(e->vtype == t_int);
			assert(e->opc == 2);
			intptr = addr_int(e->opv[0]);
			v.vtype = t_int;
			v.u.vint = ( (*intptr) *= expr_eval_int(e->opv[1]) );
			break;

		case n_adiv:		/* /= */
			assert(e->vtype == t_int);
			assert(e->opc == 2);
			assert(is_lvalue(e->opv[0]));
			intptr = addr_int(e->opv[0]);
			expr_eval( e->opv[1], &(e->vop[1]) );
			assert(e->vop[1].vtype == t_int);
			if (e->vop[1].u.vint == 0) runtime_error(div0);
			v.vtype = t_int;
			v.u.vint = ( (*intptr) /= e->vop[1].u.vint );
			break;

		case n_amod:		/* %= */
			assert(e->vtype == t_int);
			assert(e->opc == 2);
			assert(is_lvalue(e->opv[0]));
			intptr = addr_int(e->opv[0]);
			expr_eval( e->opv[1], &(e->vop[1]) );
			assert(e->vop[1].vtype == t_int);
			if (e->vop[1].u.vint == 0) runtime_error(div0);
			v.vtype = t_int;
			v.u.vint = ( (*intptr) /= e->vop[1].u.vint );
			break;

		case n_axor:		/* ^= */
			assert(e->vtype == t_int);
			assert(e->opc == 2);
			assert(is_lvalue(e->opv[0]));
			intptr = addr_int(e->opv[0]);
			v.vtype = t_int;
			v.u.vint = ( (*intptr) ^= expr_eval_int(e->opv[1]) );
			break;

		case n_aor:		/* |= */
			assert(e->vtype == t_int);
			assert(e->opc == 2);
			assert(is_lvalue(e->opv[0]));
			intptr = addr_int(e->opv[0]);
			v.vtype = t_int;
			v.u.vint = ( (*intptr) |= expr_eval_int(e->opv[1]) );
			break;

		case n_aand:		/* &= */
			assert(e->vtype == t_int);
			assert(e->opc == 2);
			intptr = addr_int(e->opv[0]);
			v.vtype = t_int;
			v.u.vint = ( (*intptr) &= expr_eval_int(e->opv[1]) );
			break;

		case n_lshft:		/* <<= */
			assert(e->vtype == t_int);
			assert(e->opc == 2);
			assert(is_lvalue(e->opv[0]));
			intptr = addr_int(e->opv[0]);
			v.vtype = t_int;
			v.u.vint = ( (*intptr) <<= expr_eval_int(e->opv[1]) );
			break;

		case n_rshft:		/* >>= */
			assert(e->vtype == t_int);
			assert(e->opc == 2);
			assert(is_lvalue(e->opv[0]));
			intptr = addr_int(e->opv[0]);
			v.vtype = t_int;
			v.u.vint = ( (*intptr) >>= expr_eval_int(e->opv[1]) );
			break;

		case n_id:
			id = e->u.id;
			assert(id != NULL);
			v.vtype = e->vtype;
			switch (id->id_kind)
			{
				case id_var:
					assert(v.vtype == id->u.v_var.vtype);
					switch (v.vtype)
					{
						case t_str:
							assert(id->u.v_var.u.vstr != NULL);
							v.u.vstr = xstrdup(id->u.v_var.u.vstr);
							break;
						case t_int:
							v.u.vint = id->u.v_var.u.vint;
							break;
					}
					break;

				case id_fct:
					expr_eval_fct(e, &v);
					break;

				case id_macro:
				case id_tab:
					/* should never reach this point */
					assert(0);
					v.vtype = t_str;
					v.u.vstr = NULL;
					break;
			}
			break;

		case n_intcst:
			assert(e->vtype == t_int);
			assert(e->opc == 0);
			v.vtype = t_int;
			v.u.vint = e->u.vint;
			break;

		case n_strcst:
			assert(e->vtype == t_str);
			assert(e->opc == 0);
			assert(e->u.vstr != NULL);
			v.vtype = t_str;
			v.u.vstr = xstrdup(e->u.vstr);
			break;
	}

	/* deallocate string operands */
	for (i = 0; i < e->opc; i++)
	{
		if (e->vop[i].vtype == t_str)
		{
			assert(e->vop[i].u.vstr != NULL);
			free(e->vop[i].u.vstr);
			e->vop[i].u.vstr = NULL;
		}
	}

	/* copy back expression result */
	if (evalue != NULL)
	{
		evalue->vtype = v.vtype;
		switch (v.vtype)
		{
			case t_str:
				evalue->u.vstr = v.u.vstr;
				break;
			case t_int:
				evalue->u.vint = v.u.vint;
				break;
		}
	}
	else if (v.vtype == t_str)
	{
		assert(v.u.vstr != NULL);
		free(v.u.vstr);
	}

	return;
}


long int *addr_int(struct expression_t *e)
{
	long int *ptr;
	struct idtable_t *idptr;
	struct avl_t *tabp;

	assert(e != NULL);
	assert(e->vtype == t_int);

	switch (e->expr_node)
	{
		case n_id:
			assert(e->opc == 0);
			idptr = e->u.id;
			assert(idptr != NULL);
			assert(idptr->id_kind == id_var);
			assert(idptr->u.v_var.vtype == t_int);
			ptr = & ( idptr->u.v_var.u.vint );
			break;

		case n_index:
			tabp = addr_tab(e, 1);
			ptr = & ( tabp->u.vint );
			break;

		default:
			/* should never reach this point */
			assert(0);
			ptr = NULL;
			break;
	}

	return ptr;
}


char **addr_str(struct expression_t *e)
{
	char **ptr;
	struct idtable_t *idptr;
	struct avl_t *tabp;

	assert(e != NULL);
	assert(e->vtype == t_str);

	switch (e->expr_node)
	{
		case n_id:
			assert(e->opc == 0);
			idptr = e->u.id;
			assert(idptr != NULL);
			assert(idptr->id_kind == id_var);
			assert(idptr->u.v_var.vtype == t_str);
			ptr = & ( idptr->u.v_var.u.vstr );
			break;

		case n_index:
			tabp = addr_tab(e, 1);
			ptr = & ( tabp->u.vstr );
			break;

		default:
			/* should never reach this point */
			assert(0);
			ptr = NULL;
			break;
	}

	return ptr;
}


struct avl_t *addr_tab(struct expression_t *e, int create)
{
	struct avl_t *v;
	int inserted;
	struct expression_t *tab;
	char *sidx;
	struct idtable_t *id;

	assert(e != NULL);
	assert(e->expr_node == n_index);
	assert(e->opc == 2);

	tab = e->opv[0];
	assert(tab != NULL);
	assert(tab->expr_node == n_id);

	id = tab->u.id;
	assert(id != NULL);
	assert(id->id_kind == id_tab);
	assert(id->u.v_tab.vtype == e->vtype);

	assert(e->opv[1] != NULL);
	expr_eval( e->opv[1], &(e->vop[1]) );

	switch (e->vop[1].vtype)
	{
		case t_str:
			assert(e->vop[1].u.vstr != NULL);
			break;

		case t_int:
			sidx = xstrdup(image_int(e->vop[1].u.vint));
			e->vop[1].vtype = t_str;
			e->vop[1].u.vstr = sidx;
			break;
	}

	v = avl_search( &(id->u.v_tab.root), e->vop[1].u.vstr, create ? &inserted : NULL);
	if (create && inserted)
	{
		assert(v != NULL);
		switch (e->vtype)
		{
			case t_str:
				v->u.vstr = xstrdup("");
				break;
			case t_int:
				v->u.vint = 0;
				break;
		}
	}

	if (create)
	{
		free( e->vop[1].u.vstr );
		e->vop[1].u.vstr = NULL;
	}

	return v;
}


char *image_int(long int n)
{
	static char image[16];

	sprintf(image, "%ld", n);
	return image;
}
