
/*
hawk (Herv's awk) v1.0
Copyright (C) by Denis Valois && Cdric Llorens, 2008.
denis.valois@laposte.fr, cedric.llorens@wanadoo.fr

This software is a computer program whose purpose is to perform
pattern scanning and processing.  This is governed by a given
template containing patterns structured in regular expressions.

This software is governed by the CeCILL license under French law and
abiding by the rules of distribution of free software.  You can  use,
modify and/or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info".

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability.

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or
data to be ensured and,  more generally, to use and operate it in the
same conditions as regards security.

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.
*/

#include <stdlib.h>
#include <stdio.h>
#include <assert.h>

#include "hawk.h"
#include "options.h"
#include "dump.h"
#include "io.h"
#include "parse.h"
#include "statement.h"
#include "ablock.h"

enum flowc_t flowc = flowc_normal;

unsigned long this_visit = 0;

void start_states(struct ablock_t *ablock);
void epsilon_closure(struct nfa_state_t *orig, struct set_t *dest);
void dfs(struct nfa_state_t *orig, struct set_t *dest);
void move(struct ablock_t *ablock);
int is_accepting(struct ablock_t *ablock);


int hawk(char *fn)
{
	FILE *fp;
	int accept;
	unsigned long i;
	enum loopc_t loopc;

	assert(fn != NULL);
	fp = io_open(fn);

	if (opt_debug) fprintf(DUMP_FP, "\nHAWK\t%s\n", io_filename);

	loopc = loopc_noloop;
	flowc = flowc_normal;
	statement_interpret( begin, &loopc);

	assert(main_ablock == NULL);
	main_ablock = ablock_new(main_nfa);

	start_states(main_ablock);
	accept = is_accepting( main_ablock );
	if (opt_debug) dump_ablock(main_ablock);

	while (flowc == flowc_normal && main_ablock->st_pres.card != 0 && io_read(fp) != NULL)
	{
		if (opt_debug) fprintf(DUMP_FP, "line\t%lu\t\t'%s'\n", io_lineno, io_line);

		/* prevent multiple executions of the same action */
		for (i = 0; i < nb_nfas; i++)
		{
			set_reset( &(nfas[i]->action_executed) );
		}

		move( main_ablock );
		accept = is_accepting( main_ablock );
		if (opt_debug) dump_ablock(main_ablock);
	}

	ablock_del(main_ablock);
	main_ablock = NULL;

	if (flowc != flowc_normal)
	{
		accept = (flowc == flowc_accept);
		flowc = flowc_normal;
	}

	loopc = loopc_noloop;

	if (accept)
	{
		statement_interpret( success, &loopc );
	}
	else
	{
		statement_interpret( failure, &loopc );
	}

	if (flowc != flowc_normal) accept = (flowc == flowc_accept);

	fflush(stdout);
	fflush(stderr);
	io_close(fp, fn);

	return accept;
}


void start_states(struct ablock_t *ablock)
{
	struct nfa_t *nfa;
	struct set_t *st_pres;

	assert(ablock != NULL);
	nfa = ablock->nfa;
	st_pres = &(ablock->st_pres);
	set_reset(st_pres);

	assert(nfa->start != NULL);
	epsilon_closure( nfa->start, st_pres );
	assert(st_pres->card != 0);

	return;
}


void epsilon_closure(struct nfa_state_t *orig, struct set_t *dest)
{
	assert(dest != NULL);

	if (orig != NULL)
	{
		/* this hack in lieu of resetting all "visited" bits to 0 */
		if (++this_visit == 0)
		{
			fprintf(stderr, "*** Implementation limit exceeded: too many DFS\n");
			exit(EXIT_FAILURE);
		}

		dfs(orig, dest);
	}

	return;
}


void dfs(struct nfa_state_t *orig, struct set_t *dest)
{
	struct nfa_state_t *sp;
	struct set_t *executed;
	enum loopc_t loopc;

	assert(orig != NULL);
	assert(dest != NULL);

	assert(orig->visited < this_visit);
	orig->visited = this_visit;

	switch (orig->state_type)
	{
		case state_1line:
		case state_accept:
			set_add(dest, orig->state_id);
			break;

		case state_epsilon:
			assert(orig->next1 != NULL || orig->next2 == NULL);
			sp = orig->next1;
			/* this hack in lieu of "if (! sp->visited)" */
			if (sp != NULL && sp->visited != this_visit) dfs(sp, dest);

			sp = orig->next2;
			/* this hack in lieu of "if (! sp->visited)" */
			if (sp != NULL && sp->visited != this_visit) dfs(sp, dest);

			if (orig->u.action != NULL)
			{
				assert(orig->nfa != NULL);
				executed = &(orig->nfa->action_executed);
				assert(orig->state_id < executed->set_size);
				if (! executed->set[orig->state_id])
				{
					loopc = loopc_noloop;
					statement_interpret( orig->u.action, &loopc );
					set_add( executed, orig->state_id );
				}
			}
			break;

		case state_conj_init:
			set_add(dest, orig->state_id);

			sp = orig->next2;
			/* this hack in lieu of "if (! sp->visited)" */
			if (sp != NULL && sp->visited != this_visit) dfs(sp, dest);
			break;

		case state_conj:
			/* should never reach this point */
			assert(0);
			break;
	}

	return;
}


void move(struct ablock_t *ablock)
{
	struct nfa_t *nfa;
	struct set_t *st_pres, *st_next;
	unsigned long id;
	struct nfa_state_t *sp;
	struct ablock_conj_t *conj, *conj_next;
	int match;
	struct ablock_t *child1, *child2;

	assert(ablock != NULL);
	assert(io_line != NULL);

	nfa = ablock->nfa;
	assert(nfa != NULL);
	st_pres = &(ablock->st_pres);
	st_next = &(ablock->st_next);
	assert(st_pres->set_size == nfa->nb_states);
	assert(st_next->set_size == nfa->nb_states);
	assert(st_pres->set != NULL);  assert(st_pres->lst != NULL);
	assert(st_next->set != NULL);  assert(st_next->lst != NULL);

	assert(st_pres->card != 0);
	set_reset( st_next );

	for (id = 0; id < st_pres->set_size; id++)
	{
		/* loop over states in increasing order, so state_conj_int is processed before state_conj */
		if (! st_pres->set[id]) continue;

		assert(id < nfa->nb_states);
		sp = nfa->state[id];
		assert(sp != NULL);

		switch (sp->state_type)
		{
			case state_1line:
				assert(sp->u.pattern != NULL);
				match = pattern_match(io_line, io_linelg, sp->u.pattern);
				if (match)
				{
					assert(sp->next1 != NULL);
					assert(sp->next2 == NULL);
					epsilon_closure( sp->next1, st_next );
				}
				break;

			case state_conj_init:
				ablock_conj_new(ablock);
				conj = ablock->tail;
				assert(conj != NULL);
				conj->conj_id = sp->state_id + 1;

				assert(sp->next1 != NULL);
				assert(sp->next1->state_id == conj->conj_id);
				set_add( st_pres, conj->conj_id );

				assert(sp->u.s.nfa1 != NULL);
				assert(sp->u.s.nfa2 != NULL);
				conj->child1 = ablock_new( sp->u.s.nfa1 );
				conj->child2 = ablock_new( sp->u.s.nfa2 );
				start_states( conj->child1 );
				start_states( conj->child2 );
				if (opt_debug)
				{
					fprintf(DUMP_FP, "NFA %p state %lu conj_init:  %p\n",
					        (void *) nfa, id, (void *) conj);
					dump_conj( ablock, conj );
				}
				break;

			case state_conj:
				/* move recursively on this conjunction activations */
				child1 = NULL;
				for (conj = ablock->head; conj != NULL; conj = conj_next)
				{
					conj_next = conj->next;
					if (conj->conj_id == sp->state_id)
					{
						child1 = conj->child1;
						child2 = conj->child2;
						assert(child1 != NULL);
						assert(child2 != NULL);
						move( child1 );
						move( child2 );

						if (is_accepting(child1) && is_accepting(child2))
						{
							/* put next1/next2 as next steps */
							assert(sp->next1 != NULL);
							epsilon_closure( sp->next1, st_next );
							epsilon_closure( sp->next2, st_next );
						}

						if (child1->st_pres.card == 0 || child2->st_pres.card == 0)
						{
							if (opt_debug)
							{
								fprintf(DUMP_FP, "NFA %p state %lu conj_del:  %p\n",
								        (void *) nfa, id, (void *) conj);
							}
							conj_next = conj->next;
							ablock_conj_del( ablock, conj );
						}
						else
						{
							/* continue conjunction */
							set_add( st_next, sp->state_id );
						}
					}
				}

				/* should have processed at least 1 activation */
				assert(child1 != NULL);
				break;

			case state_accept:
				/* noop: st_next not updated */
				break;

			case state_epsilon:
				/* should never reach this point */
				assert(0);
				break;
		}
	}

	set_swap( st_next, st_pres );

	return;
}


int is_accepting(struct ablock_t *ablock)
{
	int accepting;
	unsigned long i;
	struct nfa_t *nfa;
	struct set_t *st_pres;
	unsigned long state_id;
	struct nfa_state_t *statep;

	assert(ablock != NULL);

	nfa = ablock->nfa;
	assert(nfa != NULL);
	assert(nfa->state != NULL);

	st_pres = &(ablock->st_pres);
	assert(st_pres->lst != NULL);

	accepting = 0;
	for (i = 0; i < st_pres->card; i++)
	{
		state_id = st_pres->lst[i];
		assert(state_id < nfa->nb_states);
		statep = nfa->state[state_id];
		assert(statep != NULL);

		if (statep == nfa->accept)
		{
			assert(statep->state_type == state_accept);
			accepting = 1;
			break;
		}
		else
		{
			assert(statep->state_type != state_accept);
		}
	}

	return accepting;
}


void runtime_error(char *msg)
{
	assert(msg != NULL);

	fprintf(stderr, "hawk run time error: %s\n", msg);
	exit(EXIT_FAILURE);
}
