
/*
hawk (Herv's awk) v1.0
Copyright (C) by Denis Valois && Cdric Llorens, 2008.
denis.valois@laposte.fr, cedric.llorens@wanadoo.fr

This software is a computer program whose purpose is to perform
pattern scanning and processing.  This is governed by a given
template containing patterns structured in regular expressions.

This software is governed by the CeCILL license under French law and
abiding by the rules of distribution of free software.  You can  use,
modify and/or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info".

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability.

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or
data to be ensured and,  more generally, to use and operate it in the
same conditions as regards security.

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.
*/

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <assert.h>

#include "lexic.h"
#include "io.h"
#include "xmalloc.h"

char *template_fn = NULL;
FILE *template_fp = NULL;

int token = '\0';
union lval_t lval;

void skip_blanks(void);

#define eoln		((unsigned char *) "")
#define at_eoln		(*next_char == '\0')
static unsigned char *next_char = eoln, saved_char;
static int char_is_saved = 0;

struct rw_t
{
	char *spelling;
	int token;
};
static struct rw_t rw[] =
{
	{ "BEGIN",	TOK_BEGIN	},
	{ "DECL",	TOK_DECL	},
	{ "FAILURE",	TOK_FAILURE	},
	{ "SUCCESS",	TOK_SUCCESS	},
	{ "accept",	TOK_ACCEPT	},
	{ "break",	TOK_BREAK	},
	{ "continue",	TOK_CONTINUE	},
	{ "delete",	TOK_DELETE	},
	{ "do",		TOK_DO		},
	{ "else",	TOK_ELSE	},
	{ "exit",	TOK_EXIT	},
	{ "for",	TOK_FOR		},
	{ "forall",	TOK_FORALL	},
	{ "if",		TOK_IF		},
	{ "int",	TOK_INT		},
	{ "macro",	TOK_MACRO	},
	{ "reject",	TOK_REJECT	},
	{ "str",	TOK_STR		},
	{ "while",	TOK_WHILE	}
};
int rwcmp(const void *x, const void *y);

unsigned char decode_char(void);
static char *char_err = "illegal character";


enum lexic_mode_t lexic_mode = lexic_pattern;

void get_flag(void)
{
	skip_blanks();

	token = tolower(*next_char++);
	return;
}


void get_token(void)
{
	struct rw_t sr, *rwp;
	char *cp;
	unsigned char *bos, *eos;

	skip_blanks();
	if (token == EOF) return;

	switch (*next_char)
	{
		case 'a': case 'b': case 'c': case 'd': case 'e':
		case 'f': case 'g': case 'h': case 'i': case 'j':
		case 'k': case 'l': case 'm': case 'n': case 'o':
		case 'p': case 'q': case 'r': case 's': case 't':
		case 'u': case 'v': case 'w': case 'x': case 'y': case 'z':
		case 'A': case 'B': case 'C': case 'D': case 'E':
		case 'F': case 'G': case 'H': case 'I': case 'J':
		case 'K': case 'L': case 'M': case 'N': case 'O':
		case 'P': case 'Q': case 'R': case 'S': case 'T':
		case 'U': case 'V': case 'W': case 'X': case 'Y': case 'Z':
			lval.string = (char *) next_char;
			sr.spelling = lval.string;
			do
			{
				next_char++;
			} while (isalnum(*next_char) || *next_char == '_');

			saved_char = *next_char;
			*next_char = '\0';
			assert(! char_is_saved);
			char_is_saved = 1;

			rwp = bsearch(&sr,
			              rw,
			              sizeof(rw) / sizeof(struct rw_t),
			              sizeof(struct rw_t),
			              rwcmp);
			token = (rwp != NULL) ? rwp->token : TOK_ID;
			break;

		case '0': case '1': case '2': case '3': case '4':
		case '5': case '6': case '7': case '8': case '9':
			cp = (char *) next_char;
			do
			{
				next_char++;
			} while (isalnum(*next_char));

			saved_char = *next_char;
			*next_char = '\0';
			assert(! char_is_saved);
			char_is_saved = 1;

			lval.number = str2int(cp);
			token = TOK_INTCST;
			break;

		case ':':
			next_char++;
			switch (lexic_mode)
			{
				case lexic_pattern:
					lval.string = (char *) next_char;
					next_char = eoln;
					token = TOK_REGEX;
					break;

				case lexic_expression:
					token = ':';
					break;
			}
			break;

		case '"':
			next_char++;
			bos = eos = next_char;
			while (*next_char != '"')
			{
				*eos++ = decode_char();
			}
			*eos = '\0';
			token = TOK_STRCST;
			lval.string = xstrdup((char *) bos);
			next_char++;
			break;

		case '\'':
			next_char++;
			token = TOK_INTCST;
			lval.number = decode_char();
			if (lval.number == '\'') abend(char_err);
			if (*next_char++ != '\'') abend(char_err);
			break;

		case '+':
			next_char++;
			if (*next_char == '=')
			{
				token = TOK_AINCR;
				next_char++;
			}
			else if (*next_char == '+')
			{
				token = TOK_INCR;
				next_char++;
			}
			else
			{
				token = '+';
			}
			break;

		case '-':
			next_char++;
			if (*next_char == '=')
			{
				token = TOK_ADECR;
				next_char++;
			}
			else if (*next_char == '-')
			{
				token = TOK_DECR;
				next_char++;
			}
			else
			{
				token = '-';
			}
			break;

		case '<':
			next_char++;
			if (*next_char == '=')
			{
				token = TOK_LE;
				next_char++;
			}
			else if (*next_char == '<')
			{
				next_char++;
				if (*next_char == '=')
				{
					token = TOK_ALSH;
					next_char++;
				}
				else
				{
					token = TOK_LSHIFT;
				}
			}
			else
			{
				token = '<';
			}
			break;

		case '>':
			next_char++;
			if (*next_char == '=')
			{
				token = TOK_GE;
				next_char++;
			}
			else if (*next_char == '>')
			{
				next_char++;
				if (*next_char == '=')
				{
					token = TOK_ARSH;
					next_char++;
				}
				else
				{
					token = TOK_RSHIFT;
				}
			}
			else
			{
				token = '>';
			}
			break;

		case '=':
			next_char++;
			if (*next_char == '=')
			{
				token = TOK_EQ;
				next_char++;
			}
			else
			{
				token = '=';
			}
			break;

		case '!':
			next_char++;
			if (*next_char == '=')
			{
				token = TOK_NE;
				next_char++;
			}
			else
			{
				token = '!';
			}
			break;

		case '&':
			next_char++;
			if (*next_char == '&')
			{
				token = TOK_AND;
				next_char++;
			}
			else if (*next_char == '=')
			{
				token = TOK_AAND;
				next_char++;
			}
			else
			{
				token = '&';
			}
			break;

		case '|':
			next_char++;
			if (*next_char == '|')
			{
				token = TOK_OR;
				next_char++;
			}
			else if (*next_char == '=')
			{
				token = TOK_AOR;
				next_char++;
			}
			else
			{
				token = '|';
			}
			break;

		case '*':
			next_char++;
			if (*next_char == '=')
			{
				token = TOK_AMULT;
				next_char++;
			}
			else
			{
				token = '*';
			}
			break;

		case '/':
			next_char++;
			if (*next_char == '=')
			{
				token = TOK_ADIV;
				next_char++;
			}
			else
			{
				token = '/';
			}
			break;

		case '%':
			next_char++;
			if (*next_char == '=')
			{
				token = TOK_AMOD;
				next_char++;
			}
			else
			{
				token = '%';
			}
			break;

		case '^':
			next_char++;
			if (*next_char == '=')
			{
				token = TOK_AXOR;
				next_char++;
			}
			else
			{
				token = '^';
			}
			break;

		default:
			token = *next_char++;
			break;
	}

	return;
}


int rwcmp(const void *x, const void *y)
{
	struct rw_t *rw1, *rw2;

	rw1 = (struct rw_t *) x;
	rw2 = (struct rw_t *) y;
	assert(rw1 != NULL && rw2 != NULL);
	return strcmp(rw1->spelling, rw2->spelling);
}


void abend(char *msg)
{
	assert(msg != NULL);
	if (char_is_saved) *next_char = saved_char;

	fprintf(stderr, "*** ERROR in file '%s' line %lu, at or before ",
	        template_fn, io_lineno);
	if (at_eoln)
	{
		fprintf(stderr, "end-of-line:\n");
	}
	else if (isprint(*next_char))
	{
		fprintf(stderr, "'%c':\n", *next_char);
	}
	else
	{
		fprintf(stderr, "'\\0x%02x':\n", *next_char);
	}
	fprintf(stderr, "*** %s.\n", msg);
	exit(EXIT_FAILURE);
}


void skip_blanks(void)
{
	assert(next_char != NULL);
	if (token == EOF) abend("read past EOF");

	if (char_is_saved)
	{
		*next_char = saved_char;
		char_is_saved = 0;
	}

	do
	{
		if (at_eoln)
		{
			if ((next_char = (unsigned char *) io_read(template_fp)) == NULL)
			{
				token = EOF;
				next_char = eoln;
				return;
			}
		}
		while (isspace(*next_char)) next_char++;
		if (*next_char == '#') next_char = eoln;
	} while (at_eoln || isspace(*next_char));

	return;
}


unsigned char decode_char(void)
{
	unsigned char c;

	if (*next_char == '\\')
	{
		next_char++;
		switch (*next_char++)
		{
			case '\\':
				c = '\\';
				break;

			case '\'':
				c = '\'';
				break;

			case '"':
				c = '"';
				break;

			case '0':
				c = '\0';
				break;

			case 't':
				c = '\t';
				break;

			case 'n':
				c = '\n';
				break;

			case 'v':
				c = '\v';
				break;

			case 'f':
				c = '\f';
				break;

			case 'r':
				c = '\r';
				break;

			default:
				abend(char_err);
				c = '\0';
				break;
		}
	}
	else
	{
		if (at_eoln) abend(char_err);
		c = *next_char++;
	}

	return c;
}


long int str2int(char *str)
{
	long int n;
	char *endptr;

	assert(str != NULL);

	errno = 0;
	n = strtol(str, &endptr, 0);
	if (*endptr != '\0') abend("illegal integer constant");
	if (errno != 0) abend(strerror(errno));

	return n;
}
