
/*
hawk (Herv's awk) v1.0
Copyright (C) by Denis Valois && Cdric Llorens, 2008.
denis.valois@laposte.fr, cedric.llorens@wanadoo.fr

This software is a computer program whose purpose is to perform
pattern scanning and processing.  This is governed by a given
template containing patterns structured in regular expressions.

This software is governed by the CeCILL license under French law and
abiding by the rules of distribution of free software.  You can  use,
modify and/or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info".

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability.

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or
data to be ensured and,  more generally, to use and operate it in the
same conditions as regards security.

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.
*/

#include <stdlib.h>
#include <assert.h>

#include "parse.h"
#include "expression.h"
#include "lexic.h"
#include "xmalloc.h"

void parse_conditional_expression(struct expression_t **root);
void parse_and_expression(struct expression_t **root);
void parse_relational_expression(struct expression_t **root);
void parse_shift_expression(struct expression_t **root);
void parse_additive_expression(struct expression_t **root);
void parse_multiplicative_expression(struct expression_t **root);
void parse_unary_expression(struct expression_t **root);
void parse_primary_expression(struct expression_t **root);

int is_lvalue(struct expression_t *root);


void parse_expression(struct expression_t **root)
{
	struct expression_t *left, *right;

	assert(root != NULL);
	parse_assignment_expression( root );

	if (token == ',')
	{
		left = *root;
		*root = expr_new(n_comma);

		get_token();
		parse_expression( &right );

		assert((*root)->opc == 2);
		assert(left != NULL);
		assert(right != NULL);
		(*root)->vtype = right->vtype;
		(*root)->opv[0] = left;
		(*root)->opv[1] = right;
	}

	assert(*root != NULL);
	return;
}


void parse_assignment_expression(struct expression_t **root)
{
	struct expression_t *left, *right;

	assert(root != NULL);
	parse_conditional_expression( root );

	if (token == '=' || token == TOK_AMULT || token == TOK_ADIV || token == TOK_AMOD ||
	    token == TOK_AINCR || token == TOK_ADECR || token == TOK_ALSH || token == TOK_ARSH ||
	    token == TOK_AAND || token == TOK_AOR || token == TOK_AXOR)
	{
		left = *root;
		if (! is_lvalue(left)) abend(parse_err[15]);

		switch (token)
		{
			case '=':
				*root = expr_new(n_assignment);
				break;
			case TOK_AMULT:
				*root = expr_new(n_amul);
				break;
			case TOK_ADIV:
				*root = expr_new(n_adiv);
				break;
			case TOK_AMOD:
				*root = expr_new(n_amod);
				break;
			case TOK_AINCR:
				*root = expr_new(n_ainc);
				break;
			case TOK_ADECR:
				*root = expr_new(n_adec);
				break;
			case TOK_ALSH:
				*root = expr_new(n_lshft);
				break;
			case TOK_ARSH:
				*root = expr_new(n_rshft);
				break;
			case TOK_AAND:
				*root = expr_new(n_aand);
				break;
			case TOK_AOR:
				*root = expr_new(n_aor);
				break;
			case TOK_AXOR:
				*root = expr_new(n_axor);
				break;
			default:
				/* should never reach this point */
				assert(0);
				*root = NULL;
				break;
		}
		assert((*root)->opc == 2);

		get_token();
		parse_assignment_expression( &right );

		if (left->vtype != right->vtype) abend(parse_err[16]);
		(*root)->vtype = left->vtype;

		(*root)->opv[0] = left;
		(*root)->opv[1] = right;
	}

	assert(*root != NULL);
	return;
}


void parse_conditional_expression(struct expression_t **root)
{
	struct expression_t *expr1, *expr2, *expr3;

	assert(root != NULL);
	parse_or_expression( root );

	if (token == '?')
	{
		lexic_mode = lexic_expression;
		expr1 = *root;
		assert(expr1 != NULL);
		if (expr1->vtype != t_int) abend(parse_err[0]);
		*root = expr_new(n_cond);
		assert((*root)->opc == 3);

		get_token();
		parse_expression( &expr2 );
		assert(expr2 != NULL);
		if (token != ':') abend(parse_err[17]);
		get_token();

		parse_conditional_expression( &expr3 );
		assert(expr3 != NULL);

		if (expr2->vtype != expr3->vtype) abend(parse_err[18]);
		(*root)->vtype = expr2->vtype;

		(*root)->opv[0] = expr1;
		(*root)->opv[1] = expr2;
		(*root)->opv[2] = expr3;
		lexic_mode = lexic_pattern;
	}

	assert(*root != NULL);
	return;
}


void parse_or_expression(struct expression_t **root)
{
	struct expression_t *left, *right;

	assert(root != NULL);
	parse_and_expression( root );

	if (token == TOK_OR || token == '|' || token == '^')
	{
		left = *root;
		assert(left != NULL);
		if (left->vtype != t_int) abend(parse_err[0]);

		switch (token)
		{
			case TOK_OR:
				*root = expr_new(n_logor);
				break;
			case '|':
				*root = expr_new(n_bitor);
				break;
			case '^':
				*root = expr_new(n_xor);
				break;
			default:
				/* should never reach this point */
				assert(0);
				*root = NULL;
				break;
		}
		assert((*root)->opc == 2);

		get_token();
		parse_or_expression( &right );
		assert(right != NULL);
		if (right->vtype != t_int) abend(parse_err[0]);

		(*root)->vtype = t_int;
		(*root)->opv[0] = left;
		(*root)->opv[1] = right;
	}

	assert(*root != NULL);
	return;
}


void parse_and_expression(struct expression_t **root)
{
	struct expression_t *left, *right;

	assert(root != NULL);
	parse_relational_expression( root );

	if (token == TOK_AND || token == '&')
	{
		left = *root;
		assert(left != NULL);
		if (left->vtype != t_int) abend(parse_err[0]);

		switch (token)
		{
			case TOK_AND:
				*root = expr_new(n_logand);
				break;
			case '&':
				*root = expr_new(n_bitand);
				break;
			default:
				/* should never reach this point */
				assert(0);
				*root = NULL;
				break;
		}
		assert((*root)->opc == 2);

		get_token();
		parse_and_expression( &right );
		assert(right != NULL);
		if (right->vtype != t_int) abend(parse_err[0]);

		(*root)->vtype = t_int;
		(*root)->opv[0] = left;
		(*root)->opv[1] = right;
	}

	assert(*root != NULL);
	return;
}


void parse_relational_expression(struct expression_t **root)
{
	struct expression_t *left, *right;

	assert(root != NULL);
	parse_shift_expression( root );

	if (token == TOK_EQ || token == TOK_NE || token == '<' || token == '>' ||
	    token == TOK_LE || token == TOK_GE)
	{
		left = *root;
		assert(left != NULL);

		switch (token)
		{
			case TOK_EQ:
				*root = expr_new(n_eq);
				break;
			case TOK_NE:
				*root = expr_new(n_ne);
				break;
			case '<':
				*root = expr_new(n_lt);
				break;
			case '>':
				*root = expr_new(n_gt);
				break;
			case TOK_LE:
				*root = expr_new(n_le);
				break;
			case TOK_GE:
				*root = expr_new(n_ge);
				break;
			default:
				/* should never reach this point */
				assert(0);
				*root = NULL;
				break;
		}
		assert((*root)->opc == 2);

		get_token();
		parse_relational_expression( &right );
		assert(right != NULL);
		if (left->vtype != right->vtype) abend(parse_err[19]);

		(*root)->vtype = t_int;
		(*root)->opv[0] = left;
		(*root)->opv[1] = right;
	}

	assert(*root != NULL);
	return;
}


void parse_shift_expression(struct expression_t **root)
{
	struct expression_t *left, *right;

	assert(root != NULL);
	parse_additive_expression( root );

	if (token == TOK_LSHIFT || token == TOK_RSHIFT)
	{
		left = *root;
		assert(left != NULL);
		if (left->vtype != t_int) abend(parse_err[0]);

		switch (token)
		{
			case TOK_LSHIFT:
				*root = expr_new(n_lshift);
				break;
			case TOK_RSHIFT:
				*root = expr_new(n_rshift);
				break;
			default:
				/* should never reach this point */
				assert(0);
				*root = NULL;
				break;
		}
		assert((*root)->opc == 2);

		get_token();
		parse_shift_expression( &right );
		assert(right != NULL);
		if (right->vtype != t_int) abend(parse_err[0]);

		(*root)->vtype = t_int;
		(*root)->opv[0] = left;
		(*root)->opv[1] = right;
	}

	assert(*root != NULL);
	return;
}


void parse_additive_expression(struct expression_t **root)
{
	struct expression_t *left, *right;

	assert(root != NULL);
	parse_multiplicative_expression( root );

	if (token == '+' || token == '-')
	{
		left = *root;
		assert(left != NULL);
		if (left->vtype != t_int) abend(parse_err[0]);

		switch (token)
		{
			case '+':
				*root = expr_new(n_add);
				break;
			case '-':
				*root = expr_new(n_sub);
				break;
			default:
				/* should never reach this point */
				assert(0);
				*root = NULL;
				break;
		}
		assert((*root)->opc == 2);

		get_token();
		parse_additive_expression( &right );
		assert(right != NULL);
		if (right->vtype != t_int) abend(parse_err[0]);

		(*root)->vtype = t_int;
		(*root)->opv[0] = left;
		(*root)->opv[1] = right;
	}

	assert(*root != NULL);
	return;
}


void parse_multiplicative_expression(struct expression_t **root)
{
	struct expression_t *left, *right;

	assert(root != NULL);
	parse_unary_expression( root );

	if (token == '*' || token == '/' || token == '%')
	{
		left = *root;
		assert(left != NULL);
		if (left->vtype != t_int) abend(parse_err[0]);

		switch (token)
		{
			case '*':
				*root = expr_new(n_mul);
				break;
			case '/':
				*root = expr_new(n_div);
				break;
			case '%':
				*root = expr_new(n_mod);
				break;
			default:
				/* should never reach this point */
				assert(0);
				*root = NULL;
				break;
		}
		assert((*root)->opc == 2);

		get_token();
		parse_multiplicative_expression( &right );
		assert(right != NULL);
		if (right->vtype != t_int) abend(parse_err[0]);

		(*root)->vtype = t_int;
		(*root)->opv[0] = left;
		(*root)->opv[1] = right;
	}

	assert(*root != NULL);
	return;
}


void parse_unary_expression(struct expression_t **root)
{
	struct expression_t *right;

	assert(root != NULL);
	if (token == '+' || token == '-' || token == '~' || token == '!')
	{
		switch (token)
		{
			case '+':
				*root = expr_new(n_plus);
				break;

			case '-':
				*root = expr_new(n_minus);
				break;

			case '~':
				*root = expr_new(n_compl);
				break;

			case '!':
				*root = expr_new(n_not);
				break;

			default:
				/* should never reach this point */
				assert(0);
				*root = NULL;
				break;
		}
		assert((*root)->opc == 1);

		get_token();
		parse_unary_expression( &right );
		assert(right != NULL);
		if (right->vtype != t_int) abend(parse_err[0]);

		(*root)->vtype = t_int;
		(*root)->opv[0] = right;
	}
	else
	{
		parse_primary_expression( root );
	}

	assert(*root != NULL);
	return;
}


void parse_primary_expression(struct expression_t **root)
{
	struct idtable_t *idt;
	struct expression_t *expr;

	assert(root != NULL);
	switch (token)
	{
		case TOK_ID:
			idt = searchid(lval.string);

			*root = expr_new(n_id);
			assert((*root)->opc == 0);
			(*root)->u.id = idt;
			switch (idt->id_kind)
			{
				case id_macro:
					abend(parse_err[21]);
					break;
				case id_var:
					(*root)->vtype = idt->u.v_var.vtype;
					break;
				case id_tab:
					(*root)->vtype = idt->u.v_tab.vtype;
					break;
				case id_fct:
					(*root)->vtype = idt->u.s2.vtype;
					break;
			}
			get_token();

			if (token == '[')
			{
				if (idt->id_kind != id_tab) abend(parse_err[9]);
				expr = *root;
				*root = expr_new(n_index);
				(*root)->vtype = expr->vtype;
				assert((*root)->opc == 2);
				(*root)->opv[0] = expr;

				get_token();
				parse_expression( &expr );
				/* expr->vtype may be string or integer */
				(*root)->opv[1] = expr;

				if (token != ']') abend(parse_err[13]);
				get_token();
			}
			else if (idt->id_kind == id_fct)
			{
				parse_fct( root );
			}
			break;

		case TOK_INTCST:
			*root = expr_new(n_intcst);
			assert((*root)->opc == 0);
			(*root)->vtype = t_int;
			(*root)->u.vint = lval.number;
			get_token();
			break;

		case TOK_STRCST:
			*root = expr_new(n_strcst);
			assert((*root)->opc == 0);
			(*root)->vtype = t_str;
			(*root)->u.vstr = lval.string;
			get_token();
			break;

		case '(':
			get_token();
			parse_expression( root );
			if (token != ')') abend(parse_err[3]);
			get_token();
			break;

		case TOK_INCR:
		case TOK_DECR:
			*root = expr_new( (token == TOK_INCR) ? n_incrp : n_decrp );
			get_token();
			parse_primary_expression( &expr );
			if (expr->vtype != t_int) abend(parse_err[0]);
			if (! is_lvalue(expr)) abend(parse_err[15]);
			assert((*root)->opc == 1);
			(*root)->vtype = t_int;
			(*root)->opv[0] = expr;
			break;

		default:
			abend(parse_err[20]);
			break;
	}

	while (token == TOK_INCR || token == TOK_DECR)
	{
		expr = *root;
		if (expr->vtype != t_int) abend(parse_err[0]);
		if (! is_lvalue(expr)) abend(parse_err[15]);

		*root = expr_new( (token == TOK_INCR) ? n_incrs : n_decrs );
		assert((*root)->opc == 1);
		(*root)->vtype = t_int;
		(*root)->opv[0] = expr;

		get_token();
	}

	assert(*root != NULL);
	return;
}


int is_lvalue(struct expression_t *root)
{
	int lvalue;
	struct idtable_t *idt;

	assert(root != NULL);
	switch (root->expr_node)
	{
		case n_incrs:
		case n_decrs:
		case n_incrp:
		case n_decrp:
			lvalue = 1;
			break;

		case n_index:
			lvalue = 1;
			break;

		case n_id:
			idt = root->u.id;
			assert(idt != NULL);
			assert(idt->id_kind != id_macro && idt->id_kind != id_tab);
			lvalue = (idt->id_kind == id_var);
			break;

		default:
			lvalue = 0;
			break;
	}

	return lvalue;
}
