
/*
hawk (Herv's awk) v1.0
Copyright (C) by Denis Valois && Cdric Llorens, 2008.
denis.valois@laposte.fr, cedric.llorens@wanadoo.fr

This software is a computer program whose purpose is to perform
pattern scanning and processing.  This is governed by a given
template containing patterns structured in regular expressions.

This software is governed by the CeCILL license under French law and
abiding by the rules of distribution of free software.  You can  use,
modify and/or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info".

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability.

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or
data to be ensured and,  more generally, to use and operate it in the
same conditions as regards security.

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.
*/

#include <stdlib.h>
#include <assert.h>

#include "parse.h"
#include "atom.h"
#include "lexic.h"
#include "io.h"

#define sy_unary_op(c)		((c) == '*' || (c) == '+' || (c) == '?' || (c) == '!')
#define sy_type(c)		((c) == 'f' || (c) == 'r' || (c) == 'e')
#define sy_kase(c)		((c) == 'c' || (c) == 'i')
#define sy_line(c)		((c) == 's' || (c) == 'x')
#define sy_flag(c)		(sy_type(c) || sy_kase(c) || sy_line(c))
#define first_pattern(c)	(sy_unary_op(c) || (c) == TOK_ID || (c) == '[' || (c) == TOK_REGEX || (c) == '(')

void parse_pattern_disj(struct pattern_t **root, struct flags_t flags);
void parse_pattern_conj(struct pattern_t **root, struct flags_t flags);
void parse_pattern_concat(struct pattern_t **root, struct flags_t flags);
void parse_pattern_closure(struct pattern_t **root, struct flags_t flags);
void parse_pattern_block(struct pattern_t **root, struct flags_t flags);
void parse_flags(struct flags_t *flags);


void parse_pattern(struct pattern_t **root)
{
	struct flags_t flags;

	assert(root != NULL);

	flags.type = DEFAULT_FLAG_TYPE;
	flags.kase = DEFAULT_FLAG_KASE;
	flags.line = DEFAULT_FLAG_LINE;

	parse_pattern_disj( root, flags );
	if (*root == NULL) abend(parse_err[29]);

	pattern_normalize(*root);

	return;
}


void parse_pattern_disj(struct pattern_t **root, struct flags_t flags)
{
	struct pattern_t *child1, *child2;

	parse_pattern_conj( root, flags );

	while (token == '|')
	{
		get_token();

		child1 = *root;
		parse_pattern_conj( &child2, flags );

		*root = pattern_new(top_or);
		(*root)->u.n.child1 = child1;
		(*root)->u.n.child2 = child2;
	}

	return;
}


void parse_pattern_conj(struct pattern_t **root, struct flags_t flags)
{
	struct pattern_t *child1, *child2;

	parse_pattern_concat( root, flags );

	while (token == '&')
	{
		get_token();

		child1 = *root;
		parse_pattern_concat( &child2, flags );

		*root = pattern_new(top_and);
		(*root)->u.n.child1 = child1;
		(*root)->u.n.child2 = child2;
	}

	return;
}


void parse_pattern_concat(struct pattern_t **root, struct flags_t flags)
{
	struct pattern_t *child1, *child2;

	parse_pattern_closure( root, flags );

	while (first_pattern(token))
	{
		child1 = *root;
		parse_pattern_closure( &child2, flags );

		*root = pattern_new(top_concat);
		(*root)->u.n.child1 = child1;
		(*root)->u.n.child2 = child2;
	}

	return;
}


void parse_pattern_closure(struct pattern_t **root, struct flags_t flags)
{
	enum top_t top;
	struct pattern_t *child;

	switch (token)
	{
		case '*':
			top = top_star;
			break;

		case '+':
			top = top_plus;
			break;

		case '?':
			top = top_maybe;
			break;

		case '!':
			top = top_not;
			break;

		default:
			top = top_leaf;
			break;
	}

	if (top == top_leaf)
	{
		parse_pattern_block( root, flags );
		if (token == '{')
		{
			if ((*root)->action != NULL) abend("action already defined");
			parse_statement( &((*root)->action) );
		}
	}
	else
	{
		get_token();
		parse_pattern_closure( &child, flags );

		*root = pattern_new(top);
		(*root)->u.n.child1 = child;
	}

	return;
}


void parse_pattern_block(struct pattern_t **root, struct flags_t flags)
{
	struct idtable_t *patp;

	if (token == TOK_ID)
	{
		patp = searchid(lval.string);
		if (patp->id_kind != id_macro) abend(parse_err[30]);
		if (patp->u.s1.v_pat == NULL) abend(parse_err[30]);
		assert(patp->u.s1.v_stat == NULL);

		*root = pattern_cpy(patp->u.s1.v_pat);
		get_token();
	}
	else
	{
		if (token == '[')
		{
			parse_flags( &flags );
			if (token != ']') abend(parse_err[13]);
			get_token();
		}

		switch (token)
		{
			case TOK_REGEX:
				*root = pattern_new(top_leaf);
				(*root)->u.ap = atom_new(flags, io_lineno, lval.string);
				get_token();
				break;

			case '(':
				get_token();
				parse_pattern_disj( root, flags );
				if (token != ')') abend(parse_err[3]);
				get_token();
				break;

			default:
				abend(parse_err[29]);
				break;
		}
	}

	return;
}


void parse_flags(struct flags_t *flags)
{
	int type_given, kase_given, line_given;

	assert(flags != NULL);
	type_given = 0; kase_given = 0; line_given = 0;

	assert(token == '[');
	get_flag();

	while (sy_flag(token))
	{
		switch (token)
		{
			case 'f':
			case 'r':
			case 'e':
				if (type_given) abend("duplicated flags 'fre'");
				type_given = 1;
				break;

			case 'c':
			case 'i':
				if (kase_given) abend("duplicated flags 'ci'");
				kase_given = 1;
				break;

			case 's':
			case 'x':
				if (line_given) abend("duplicated flags 'sx'");
				line_given = 1;
				break;

			default:
				/* should never reach this point */
				assert(0);
				break;
		}

		switch (token)
		{
			case 'f':
				flags->type = flag_f;
				break;
			case 'r':
				flags->type = flag_r;
				break;
			case 'e':
				flags->type = flag_e;
				break;
			case 'c':
				flags->kase = flag_c;
				break;
			case 'i':
				flags->kase = flag_i;
				break;
			case 's':
				flags->line = flag_s;
				break;
			case 'x':
				flags->line = flag_x;
				break;
			default:
				/* should never reach this point */
				assert(0);
				break;
		}

		get_flag();
	}

	return;
}
