
/*
hawk (Herv's awk) v1.0
Copyright (C) by Denis Valois && Cdric Llorens, 2008.
denis.valois@laposte.fr, cedric.llorens@wanadoo.fr

This software is a computer program whose purpose is to perform
pattern scanning and processing.  This is governed by a given
template containing patterns structured in regular expressions.

This software is governed by the CeCILL license under French law and
abiding by the rules of distribution of free software.  You can  use,
modify and/or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info".

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability.

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or
data to be ensured and,  more generally, to use and operate it in the
same conditions as regards security.

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.
*/

#include <stdlib.h>
#include <stdio.h>
#include <assert.h>

#include "pattern.h"
#include "lexic.h"
#include "xmalloc.h"

unsigned long lineof(struct pattern_t *root);


struct pattern_t *pattern_new(enum top_t top)
{
	struct pattern_t *tp;

	tp = xmalloc(sizeof(struct pattern_t));
	tp->top = top;

	if (top == top_leaf)
	{
		tp->nullable = 0;
		tp->pattern1 = 1;
		tp->u.ap = NULL;
	}
	else
	{
		tp->nullable = 1;
		tp->pattern1 = 0;
		tp->u.n.child1 = NULL;
		tp->u.n.child2 = NULL;
	}

	tp->action = NULL;

	return tp;
}


struct pattern_t *pattern_cpy(struct pattern_t *root)
{
	struct pattern_t *tp;

	assert(root != NULL);
	tp = xmalloc(sizeof(struct pattern_t));

	tp->top = root->top;
	tp->nullable = root->nullable;
	tp->pattern1 = root->pattern1;
	tp->action = root->action;
	if (tp->action != NULL) tp->action->ref_cnt++;

	switch (root->top)
	{
		case top_or:
		case top_and:
		case top_concat:
			assert(root->u.n.child2 != NULL);
			tp->u.n.child2 = pattern_cpy(root->u.n.child2);
			/* fall through */

		case top_not:
		case top_star:
		case top_plus:
		case top_maybe:
			assert(root->u.n.child1 != NULL);
			tp->u.n.child1 = pattern_cpy(root->u.n.child1);
			break;

		case top_leaf:
			assert(root->u.ap != NULL);
			tp->u.ap = root->u.ap;
			tp->u.ap->ref_cnt++;
			break;
	}

	return tp;
}


void pattern_normalize(struct pattern_t *root)
{
	struct pattern_t *child1 = NULL, *child2 = NULL;

	assert(root != NULL);

	/* postfix walk */
	switch (root->top)
	{
		case top_or:
		case top_and:
		case top_concat:
			child2 = root->u.n.child2;
			assert(child2 != NULL);
			pattern_normalize( child2 );
			/* fall through */

		case top_not:
		case top_star:
		case top_plus:
		case top_maybe:
			child1 = root->u.n.child1;
			assert(child1 != NULL);
			pattern_normalize( child1 );
			break;

		case top_leaf:
			/* noop */
			assert(root->u.ap != NULL);
			break;
	}

	/* computes "nullable" */
	switch (root->top)
	{
		case top_not:
			root->nullable = child1->nullable;
			break;

		case top_plus:
			root->nullable = child1->nullable;
			break;

		case top_or:
			root->nullable = child1->nullable || child2->nullable;
			break;

		case top_and:
		case top_concat:
			root->nullable = child1->nullable && child2->nullable;
			break;

		case top_star:
		case top_maybe:
			root->nullable = 1;
			break;

		case top_leaf:
			root->nullable = 0;
			break;
	}

	if (root->nullable && root->action != NULL)
	{
		fprintf(stderr, "file '%s' at or before line %lu:\n",
		        template_fn, lineof(child1));
		fprintf(stderr, "action associated to a nullable pattern.\n");
		exit(EXIT_FAILURE);
	}

	/* determines if root is fast-evaluable (pattern1) */
	switch (root->top)
	{
		case top_not:
			root->pattern1 = child1->pattern1;
			if (! root->pattern1)
			{
				fprintf(stderr, "file '%s' at or before line %lu:\n",
				        template_fn, lineof(child1));
				fprintf(stderr, "operator '!' applied on a complex pattern.\n");
				exit(EXIT_FAILURE);
			}
			break;

		case top_or:
		case top_and:
			root->pattern1 = child1->pattern1 && child2->pattern1;
			break;

		case top_star:
		case top_plus:
		case top_maybe:
		case top_concat:
			root->pattern1 = 0;
			break;

		case top_leaf:
			root->pattern1 = 1;
			break;
	}

	return;
}


int pattern_match(char *line, size_t lg, struct pattern_t *root)
{
	int r, r1, r2;
	struct atom_t *ap;
	enum loopc_t loopc;

	assert(line != NULL);
	assert(root != NULL);
	assert(root->pattern1);

	switch (root->top)
	{
		case top_not:
			r = ! pattern_match(line, lg, root->u.n.child1);
			break;

		case top_or:
			/* no short-circuit; interpret children's action */
			r1 = pattern_match(line, lg, root->u.n.child1);
			r2 = pattern_match(line, lg, root->u.n.child2);
			r = r1 || r2;
			break;

		case top_and:
			/* no short-circuit; interpret children's action */
			r1 = pattern_match(line, lg, root->u.n.child1);
			r2 = pattern_match(line, lg, root->u.n.child2);
			r = r1 && r2;
			break;

		case top_leaf:
			ap = root->u.ap;
			assert(ap != NULL);
			r = atom_match(line, lg, ap);
			break;

		default:
			/* should never reach this point */
			assert(0);
			r = 0;
			break;
	}

	if (r && root->action != NULL)
	{
		loopc = loopc_noloop;
		statement_interpret( root->action, &loopc );
	}

	return r;
}


void pattern_free(struct pattern_t *root)
{
	if (root != NULL)
	{
		statement_free( &(root->action) );

		if (root->top == top_leaf)
		{
			atom_free( &(root->u.ap) );
		}
		else
		{
			pattern_free( root->u.n.child1 );
			pattern_free( root->u.n.child2 );
		}

		free( root );
	}

	return;
}


unsigned long lineof(struct pattern_t *root)
{
	struct atom_t *ap;

	assert(root != NULL);
	while (root->top != top_leaf)
	{
		root = root->u.n.child1;
		assert(root != NULL);
	}

	ap = root->u.ap;
	assert(ap != NULL);
	return ap->lineno;
}
