
/*
hawk (Herv's awk) v1.0
Copyright (C) by Denis Valois && Cdric Llorens, 2008.
denis.valois@laposte.fr, cedric.llorens@wanadoo.fr

This software is a computer program whose purpose is to perform
pattern scanning and processing.  This is governed by a given
template containing patterns structured in regular expressions.

This software is governed by the CeCILL license under French law and
abiding by the rules of distribution of free software.  You can  use,
modify and/or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info".

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability.

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or
data to be ensured and,  more generally, to use and operate it in the
same conditions as regards security.

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.
*/

#include <stdlib.h>
#include <stdio.h>
#include <assert.h>

#include "statement.h"
#include "xmalloc.h"
#include "avl.h"
#include "hawk.h"

void stat_delete(struct expression_t *expr);
void stat_forall(struct avl_t *t_root, enum vtype_t t_vtype,
                 struct evalue_t *t_index, struct evalue_t *t_value,
                 struct statement_t *t_body, enum loopc_t *loopc);


struct statement_t *statement_new(enum st_type_t st_type)
{
	struct statement_t *sp;

	sp = xmalloc(sizeof(struct statement_t));

	sp->ref_cnt = 1;
	sp->st_type = st_type;

	switch (st_type)
	{
		case st_list:
			sp->u.p_list = NULL;
			break;

		case st_expression:
			sp->u.p_expression = NULL;
			break;

		case st_if:
			sp->u.p_if.expr = NULL;
			sp->u.p_if.p_then = NULL;
			sp->u.p_if.p_else = NULL;
			break;

		case st_while:
		case st_do:
			sp->u.p_while.expr = NULL;
			sp->u.p_while.p_body = NULL;
			break;

		case st_for:
			sp->u.p_for.expr1 = NULL;
			sp->u.p_for.expr2 = NULL;
			sp->u.p_for.expr3 = NULL;
			sp->u.p_for.p_body = NULL;
			break;

		case st_forall:
			sp->u.p_forall.p_array = NULL;
			sp->u.p_forall.p_index = NULL;
			sp->u.p_forall.p_value = NULL;
			sp->u.p_forall.p_body = NULL;
			break;

		case st_break:
		case st_continue:
		case st_accept:
		case st_reject:
			break;

		case st_delete:
		case st_exit:
			sp->u.p_delete.expr1 = NULL;
			break;
	}

	return sp;
}


void statement_interpret(struct statement_t *st, enum loopc_t *loopc)
{
	enum loopc_t this_loop;
	struct stat_list_t *slp;
	struct expression_t *expr1, *expr2, *expr3;

	assert(loopc != NULL);
	if (st != NULL && flowc == flowc_normal && (*loopc == loopc_noloop || *loopc == loopc_inloop))
	{
		switch (st->st_type)
		{
			case st_list:
				for (slp = st->u.p_list; slp != NULL; slp = slp->next)
				{
					statement_interpret(slp->sp, loopc);
					if (*loopc == loopc_break || *loopc == loopc_continue) break;
					if (flowc != flowc_normal) break;
				}
				break;

			case st_expression:
				expr_eval(st->u.p_expression, NULL);
				break;

			case st_if:
				if (expr_eval_int(st->u.p_if.expr) != 0)
				{
					statement_interpret(st->u.p_if.p_then, loopc);
				}
				else
				{
					statement_interpret(st->u.p_if.p_else, loopc);
				}
				break;

			case st_while:
				while (expr_eval_int(st->u.p_while.expr) != 0)
				{
					this_loop = loopc_inloop;
					statement_interpret(st->u.p_while.p_body, &this_loop);
					assert(this_loop != loopc_noloop);
					if (this_loop == loopc_break) break;
					if (flowc != flowc_normal) break;
				}
				break;

			case st_do:
				do
				{
					this_loop = loopc_inloop;
					statement_interpret(st->u.p_while.p_body, &this_loop);
					assert(this_loop != loopc_noloop);
					if (this_loop == loopc_break) break;
					if (flowc != flowc_normal) break;
				}
				while (expr_eval_int(st->u.p_while.expr) != 0);
				break;

			case st_for:
				expr1 = st->u.p_for.expr1;
				expr2 = st->u.p_for.expr2;
				expr3 = st->u.p_for.expr3;

				if (expr1 != NULL) expr_eval(expr1, NULL);

				while ( (expr2 == NULL) || expr_eval_int(expr2) )
				{
					this_loop = loopc_inloop;
					statement_interpret(st->u.p_for.p_body, &this_loop);
					assert(this_loop != loopc_noloop);
					if (this_loop == loopc_break) break;
					if (flowc != flowc_normal) break;

					if (expr3 != NULL) expr_eval(expr3, NULL);
				}
				break;

			case st_forall:
				assert(st->u.p_forall.p_array != NULL);
				assert(st->u.p_forall.p_index != NULL);
				assert(st->u.p_forall.p_value != NULL);
				assert(st->u.p_forall.p_array->id_kind == id_tab);
				assert(st->u.p_forall.p_index->id_kind == id_var);
				assert(st->u.p_forall.p_value->id_kind == id_var);
				assert(st->u.p_forall.p_index->u.v_var.vtype == t_str);
				assert(st->u.p_forall.p_array->u.v_var.vtype == st->u.p_forall.p_value->u.v_var.vtype);

				this_loop = loopc_inloop;
				stat_forall(st->u.p_forall.p_array->u.v_tab.root,
				            st->u.p_forall.p_array->u.v_tab.vtype,
				            &(st->u.p_forall.p_index->u.v_var),
				            &(st->u.p_forall.p_value->u.v_var),
				            st->u.p_forall.p_body, &this_loop);
				break;

			case st_break:
				*loopc = loopc_break;
				break;

			case st_continue:
				*loopc = loopc_continue;
				break;

			case st_delete:
				stat_delete(st->u.p_delete.expr1);
				break;

			case st_exit:
				exit ((int) expr_eval_int(st->u.p_delete.expr1));
				break;

			case st_accept:
				flowc = flowc_accept;
				break;

			case st_reject:
				flowc = flowc_reject;
				break;
		}
	}

	return;
}


void stat_delete(struct expression_t *expr)
{
	struct idtable_t *tabp;
	struct evalue_t tab_idx;
	char *sidx;
	struct avl_t *tab_el;

	assert(expr != NULL);
	switch (expr->expr_node)
	{
		case n_id:
			tabp = expr->u.id;
			assert(tabp != NULL);
			assert(tabp->id_kind == id_tab);
			avl_delete( &(tabp->u.v_tab.root), tabp->u.v_tab.vtype);
			break;

		case n_index:
			assert(expr->opc == 2);
			assert(expr->opv[0] != NULL);
			assert(expr->opv[0]->expr_node == n_id);
			tabp = expr->opv[0]->u.id;
			assert(tabp != NULL);
			assert(tabp->id_kind == id_tab);

			assert(expr->opv[1] != NULL);
			expr_eval(expr->opv[1], &tab_idx);
			switch (tab_idx.vtype)
			{
				case t_str:
					assert(tab_idx.u.vstr != NULL);
					break;
				case t_int:
					sidx = xstrdup(image_int(tab_idx.u.vint));
					tab_idx.vtype = t_str;
					tab_idx.u.vstr = sidx;
					break;
			}

			tab_el = avl_remove( &(tabp->u.v_tab.root), tab_idx.u.vstr );
			free(tab_idx.u.vstr);
			if (tab_el != NULL && tabp->u.v_tab.vtype == t_str)
			{
				assert(tab_el->u.vstr != NULL);
				free(tab_el->u.vstr);
			}
			break;

		default:
			/* should never reach this point */
			assert(0);
			break;
	}

	return;
}


void stat_forall(struct avl_t *t_root, enum vtype_t t_vtype,
                 struct evalue_t *t_index, struct evalue_t *t_value,
                 struct statement_t *t_body, enum loopc_t *loopc)
{
	assert(loopc != NULL);

	if (t_root != NULL && *loopc != loopc_break && flowc == flowc_normal)
	{
		stat_forall(t_root->left, t_vtype, t_index, t_value, t_body, loopc);
		if (*loopc == loopc_break) return;
		if (flowc != flowc_normal) return;

		assert(t_index != NULL);
		assert(t_value != NULL);

		assert(t_index->vtype == t_str);
		assert(t_index->u.vstr != NULL);
		free(t_index->u.vstr);
		t_index->u.vstr = xstrdup(t_root->key);

		assert(t_vtype == t_value->vtype);
		switch (t_vtype)
		{
			case t_str:
				assert(t_value->u.vstr != NULL);
				free(t_value->u.vstr);
				t_value->u.vstr = xstrdup(t_root->u.vstr);
				break;

			case t_int:
				t_value->u.vint = t_root->u.vint;
				break;
		}

		*loopc = loopc_inloop;
		statement_interpret(t_body, loopc);
		assert(*loopc != loopc_noloop);
		if (flowc != flowc_normal) return;

		stat_forall(t_root->right, t_vtype, t_index, t_value, t_body, loopc);
		/* if (*loopc == loopc_break) return; */
		/* if (flowc != flowc_normal) return; */
	}

	return;
}


void statement_free(struct statement_t **rootp)
{
	struct statement_t *root;
	struct stat_list_t *sl, *sn;

	assert(rootp != NULL);
	root = *rootp;

	if (root != NULL && --(root->ref_cnt) == 0)
	{
		switch (root->st_type)
		{
			case st_list:
				for (sl = root->u.p_list; sl != NULL; sl = sn)
				{
					sn = sl->next;
					statement_free( &(sl->sp) );
					free( sl );
				}
				break;

			case st_expression:
				expr_free( root->u.p_expression );
				break;

			case st_if:
				expr_free( root->u.p_if.expr );
				statement_free( &(root->u.p_if.p_then) );
				statement_free( &(root->u.p_if.p_else) );
				break;

			case st_while:
			case st_do:
				expr_free( root->u.p_while.expr );
				statement_free( &(root->u.p_while.p_body) );
				break;

			case st_for:
				expr_free( root->u.p_for.expr1 );
				expr_free( root->u.p_for.expr2 );
				expr_free( root->u.p_for.expr3 );
				statement_free( &(root->u.p_for.p_body) );
				break;

			case st_forall:
				statement_free( &(root->u.p_forall.p_body) );
				break;

			case st_break:
			case st_continue:
			case st_accept:
			case st_reject:
				/* noop */
				break;

			case st_delete:
			case st_exit:
				expr_free( root->u.p_delete.expr1 );
				break;
		}
		free( root );
		(*rootp) = NULL;
	}

	return;
}
